from gen_epix.casedb.domain import command
from gen_epix.casedb.domain.enum import Role
from gen_epix.commondb.domain.enum import Role as CommonRole
from gen_epix.commondb.domain.policy import RoleGenerator as CommonRoleGenerator
from gen_epix.fastapp import PermissionTypeSet
from gen_epix.fastapp.services.rbac import BaseRbacService


class RoleGenerator(CommonRoleGenerator):

    COMMON_ROLE_ENUM_MAP = {x: Role[x.name] for x in CommonRole}

    EXTRA_ROLE_SET_MAP = {}

    COMMON_ROLE_PERMISSION_SETS = (
        CommonRoleGenerator.map_from_common_role_permission_sets(
            COMMON_ROLE_ENUM_MAP, command.COMMON_COMMAND_MAP
        )
    )

    ROLE_PERMISSION_SETS = {
        # TODO: remove UPDATE from association objects that do not have properties of their own such as CaseTypeSetMember
        Role.APP_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.APP_ADMIN]
        | {
            # case
            (
                command.CaseSetCrudCommand,
                PermissionTypeSet.C,
            ),  # Other users can only use dedicated command
            (
                command.CaseCrudCommand,
                PermissionTypeSet.C,
            ),  # Other users can only use dedicated command
            (command.CaseTypeCrudCommand, PermissionTypeSet.D),
            (command.CaseSetCategoryCrudCommand, PermissionTypeSet.CU),
            (command.CaseSetStatusCrudCommand, PermissionTypeSet.CU),
            (command.CaseTypeSetCategoryCrudCommand, PermissionTypeSet.D),
            (command.CaseTypeSetCrudCommand, PermissionTypeSet.D),
            (
                command.DataCollectionSetDataCollectionUpdateAssociationCommand,
                PermissionTypeSet.E,
            ),
            # abac
            (command.OrganizationAccessCasePolicyCrudCommand, PermissionTypeSet.CUD),
            (
                command.OrganizationShareCasePolicyCrudCommand,
                PermissionTypeSet.CUD,
            ),
        },
        Role.REFDATA_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.REFDATA_ADMIN]
        | {
            # case
            (command.GeneticDistanceProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeColCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeColSetCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeColSetMemberCrudCommand, PermissionTypeSet.CRUD),
            (command.CaseTypeCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeSetCaseTypeUpdateAssociationCommand, PermissionTypeSet.E),
            (
                command.CaseTypeColSetCaseTypeColUpdateAssociationCommand,
                PermissionTypeSet.E,
            ),
            (command.CaseTypeSetCategoryCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeSetCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeSetMemberCrudCommand, PermissionTypeSet.CRUD),
            (command.ColCrudCommand, PermissionTypeSet.CRU),
            (command.DimCrudCommand, PermissionTypeSet.CRU),
            # ontology
            (command.ConceptCrudCommand, PermissionTypeSet.CRU),
            (command.ConceptSetCrudCommand, PermissionTypeSet.CRU),
            (command.ConceptRelationCrudCommand, PermissionTypeSet.CRU),
            (command.DiseaseCrudCommand, PermissionTypeSet.CRU),
            (
                command.DiseaseEtiologicalAgentUpdateAssociationCommand,
                PermissionTypeSet.E,
            ),
            (command.EtiologicalAgentCrudCommand, PermissionTypeSet.CRU),
            (command.EtiologyCrudCommand, PermissionTypeSet.CRU),
            (command.RegionCrudCommand, PermissionTypeSet.CRU),
            (command.RegionSetCrudCommand, PermissionTypeSet.CRU),
            (command.RegionSetShapeCrudCommand, PermissionTypeSet.CRUD),
        },
        Role.ORG_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.ORG_ADMIN]
        | {
            # abac
            (command.UserAccessCasePolicyCrudCommand, PermissionTypeSet.CUD),
            (command.UserShareCasePolicyCrudCommand, PermissionTypeSet.CUD),
        },
        Role.ORG_USER: COMMON_ROLE_PERMISSION_SETS[Role.ORG_USER]
        | {
            # case
            (command.CaseTypeColCrudCommand, PermissionTypeSet.R),
            (command.ColCrudCommand, PermissionTypeSet.R),
            (command.GeneticDistanceProtocolCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmClassCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmCrudCommand, PermissionTypeSet.R),
            (command.CaseSetCategoryCrudCommand, PermissionTypeSet.R),
            (command.CaseSetDataCollectionLinkCrudCommand, PermissionTypeSet.CRUD),
            (command.CaseSetStatusCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeColSetCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeSetCategoryCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeSetCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeColSetMemberCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeSetMemberCrudCommand, PermissionTypeSet.R),
            (
                command.CaseDataCollectionLinkCrudCommand,
                PermissionTypeSet.CRUD,
            ),
            (command.CreateCaseSetCommand, PermissionTypeSet.E),
            (command.CreateCasesCommand, PermissionTypeSet.E),
            (command.CaseSetCrudCommand, PermissionTypeSet.RUD),
            (command.CaseSetMemberCrudCommand, PermissionTypeSet.CRUD),
            (command.RetrieveCaseSetStatsCommand, PermissionTypeSet.E),
            (command.RetrieveCaseTypeStatsCommand, PermissionTypeSet.E),
            (command.RetrieveCasesByIdCommand, PermissionTypeSet.E),
            (command.RetrieveCasesByQueryCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteCaseTypeCommand, PermissionTypeSet.E),
            (command.RetrieveCaseSetRightsCommand, PermissionTypeSet.E),
            (command.RetrieveCaseRightsCommand, PermissionTypeSet.E),
            (command.RetrieveLibraryPrepProtocolsCommand, PermissionTypeSet.E),
            (command.RetrieveAssemblyProtocolsCommand, PermissionTypeSet.E),
            (command.ValidateCasesCommand, PermissionTypeSet.E),
            (command.CreateReadSetsForCasesCommand, PermissionTypeSet.E),
            (command.CreateSeqsForCasesCommand, PermissionTypeSet.E),
            (command.CreateFileForReadSetCommand, PermissionTypeSet.E),
            (command.CreateFileForSeqCommand, PermissionTypeSet.E),
            # ontology
            (command.ConceptCrudCommand, PermissionTypeSet.R),
            (command.ConceptSetCrudCommand, PermissionTypeSet.R),
            (command.DiseaseCrudCommand, PermissionTypeSet.R),
            (command.EtiologicalAgentCrudCommand, PermissionTypeSet.R),
            (command.EtiologyCrudCommand, PermissionTypeSet.R),
            (command.RegionSetCrudCommand, PermissionTypeSet.R),
            (command.RegionSetShapeCrudCommand, PermissionTypeSet.R),
            (command.RegionCrudCommand, PermissionTypeSet.R),
            # abac
            (command.OrganizationAccessCasePolicyCrudCommand, PermissionTypeSet.R),
            (command.OrganizationShareCasePolicyCrudCommand, PermissionTypeSet.R),
            (command.UserAccessCasePolicyCrudCommand, PermissionTypeSet.R),
            (command.UserShareCasePolicyCrudCommand, PermissionTypeSet.R),
            # seq
            (command.RetrieveAlleleProfileCommand, PermissionTypeSet.E),
            (command.RetrieveGeneticSequenceByCaseCommand, PermissionTypeSet.E),
            (command.RetrieveGeneticSequenceFastaByCaseCommand, PermissionTypeSet.E),
            (command.RetrievePhylogeneticTreeByCasesCommand, PermissionTypeSet.E),
            (command.RetrievePhylogeneticTreeBySequencesCommand, PermissionTypeSet.E),
        },
        Role.GUEST: COMMON_ROLE_PERMISSION_SETS[Role.GUEST] | set(),
    }

    ROLE_HIERARCHY = CommonRoleGenerator.map_from_common_role_hierarchy(
        COMMON_ROLE_ENUM_MAP
    )

    ROLE_PERMISSIONS = BaseRbacService.expand_hierarchical_role_permissions(
        ROLE_HIERARCHY, ROLE_PERMISSION_SETS  # type: ignore[arg-type]
    )



from datetime import timedelta
from typing import Any

from cachetools import TTLCache, cached

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, model
from gen_epix.commondb.domain.service.organization import BaseOrganizationService
from gen_epix.fastapp import Command, CrudOperation, exc
from gen_epix.fastapp.app import App
from gen_epix.fastapp.enum import CrudOperationSet
from gen_epix.fastapp.model import CrudCommand


class OrganizationService(BaseOrganizationService):
    DEFAULT_CFG = {
        "user_invitation_time_to_live": 86400,  # 1 day
    }

    CACHE_INVALIDATION_COMMANDS: tuple[type[Command], ...] = (
        command.UserCrudCommand,
        command.UpdateUserCommand,
    )

    def __init__(
        self,
        app: App,
        **kwargs: Any,
    ) -> None:
        super().__init__(app, **kwargs)
        app_impl: AppImplDetails = app.impl
        self.user_class: type[model.User] = app_impl.get_mapped_class(model.User)
        self.user_invitation_class: type[model.UserInvitation] = (
            app_impl.get_mapped_class(model.UserInvitation)
        )
        self.user_invitation_constraints_class: type[
            model.UserInvitationConstraints
        ] = app_impl.get_mapped_class(model.UserInvitationConstraints)

    def crud(
        self,
        cmd: CrudCommand,
    ) -> Any:
        # ABAC: ROOT may not delete self or own organization
        if (
            issubclass(
                type(cmd), (command.UserCrudCommand, command.OrganizationCrudCommand)
            )
            and cmd.operation in CrudOperationSet.DELETE.value
            and cmd.user
            and self.app.user_manager.is_root_user(cmd.user)
        ):
            user: model.User = cmd.user  # type: ignore[assignment]
            is_delete_user = issubclass(type(cmd), command.UserCrudCommand)
            id_ = user.id if is_delete_user else user.organization_id
            raise_error = False
            if cmd.operation == CrudOperation.DELETE_ALL:
                raise_error = True
            elif cmd.operation == CrudOperation.DELETE_ONE:
                raise_error = id_ == cmd.obj_ids
            elif cmd.operation == CrudOperation.DELETE_SOME:
                raise_error = id_ in cmd.obj_ids  # type: ignore[operator]
            else:
                raise NotImplementedError(
                    f"Unsupported delete operation: {cmd.operation}"
                )
            if raise_error:
                raise exc.UnauthorizedAuthError(
                    f"Root user may not delete {'self' if is_delete_user else 'own organization'}"
                )
            pass

        retval = super().crud(cmd)
        # Invalidate cache
        if issubclass(type(cmd), OrganizationService.CACHE_INVALIDATION_COMMANDS):
            self.retrieve_user_by_key.cache_clear()  # type:ignore[attr-defined]
        return retval

    @cached(cache=TTLCache(maxsize=1000, ttl=60))
    def retrieve_user_by_key(self, user_key: str) -> model.User:
        with self.repository.uow() as uow:
            return self.repository.retrieve_user_by_key(uow, user_key)

    def invite_user(
        self,
        cmd: command.InviteUserCommand,
    ) -> model.UserInvitation:
        user = cmd.user
        if user is None:
            raise exc.UnauthorizedAuthError("Command has no user")
        if user.id is None:
            raise exc.UnauthorizedAuthError("User has no ID")
        key = cmd.key
        initial_roles = cmd.roles
        organization_id = cmd.organization_id

        with self.repository.uow() as uow:
            # Verify if user already exists
            is_existing_user = self.app.user_manager.is_existing_user_by_key(
                cmd.key, uow
            )

            if is_existing_user:
                if self._logger:
                    self._logger.info(
                        self.create_log_message(
                            "acba1a0e",
                            f"User {key} already exists",
                        )
                    )
                raise exc.UserAlreadyExistsAuthError("User already exists")

            is_existing_organization = self.repository.crud(
                uow,
                user.id,
                model.Organization,
                None,
                organization_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                if self._logger:
                    self._logger.info(
                        self.create_log_message(
                            "cdf1b633",
                            f"Organization id {organization_id} does not exist",
                        )
                    )
                raise exc.InvalidIdsError("Organization does not exist")

            # Verify if invitation(s) already exists for this key, and delete those
            # TODO: Must be done within the same session to be safe,
            # so requires specific repository method
            user_invitations: list[model.UserInvitation] = self.repository.crud(
                uow,
                user.id,
                self.user_invitation_class,
                None,
                None,
                CrudOperation.READ_ALL,
            )  # type: ignore
            user_invitations = [x for x in user_invitations if x.key == key]
            if user_invitations:
                self.repository.crud(
                    uow,
                    user.id,
                    self.user_invitation_class,
                    None,
                    [x.id for x in user_invitations],
                    CrudOperation.DELETE_SOME,
                )
            # Create user invitation
            user_invitation = self.user_invitation_class(
                id=self.generate_id(),  # type: ignore[arg-type]
                key=key,
                roles=initial_roles,
                organization_id=organization_id,
                invited_by_user_id=user.id,
                token=self.generate_user_invitation_token(),
                expires_at=self.generate_timestamp()
                + timedelta(
                    seconds=self.props.get(
                        "user_invitation_time_to_live",
                        OrganizationService.DEFAULT_CFG["user_invitation_time_to_live"],
                    )
                ),
            )
            user_invitation_in_db: model.UserInvitation = self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                self.user_invitation_class,
                user_invitation,
                None,
                CrudOperation.CREATE_ONE,
            )
        return user_invitation_in_db

    def retrieve_invite_user_constraints(
        self, cmd: command.RetrieveInviteUserConstraintsCommand
    ) -> model.UserInvitationConstraints:
        sub_cmd = command.RetrieveOrganizationsUnderAdminCommand(user=cmd.user)
        sub_cmd._policies = cmd._policies
        organization_ids = self.app.handle(sub_cmd)
        roles = self.app.handle(command.RetrieveSubRolesCommand(user=cmd.user))
        return self.user_invitation_constraints_class(
            organization_ids=organization_ids,
            roles=roles,
        )

    def register_invited_user(
        self, cmd: command.RegisterInvitedUserCommand
    ) -> model.User:
        new_user = cmd.user
        if new_user is None:
            # Should not happen
            raise AssertionError("Command has no user")
        if not self.app.user_manager:
            raise exc.InvalidArgumentsError("User manager not set")

        with self.repository.uow() as uow:
            # Get possible user invitations
            user_invitations: list[model.UserInvitation] = self.repository.crud(  # type: ignore[assignment]
                uow,
                None,
                self.user_invitation_class,
                None,
                None,
                CrudOperation.READ_ALL,
            )
            now = self.generate_timestamp()
            user_invitations = [
                x
                for x in user_invitations
                if x.key == new_user.key and x.expires_at > now
            ]
            if not user_invitations:
                raise exc.UnauthorizedAuthError(
                    f"No valid invitations found for user {new_user.key}",
                )
            user_invitations_with_token = [
                x for x in user_invitations if x.token == cmd.token
            ]
            if not user_invitations_with_token:
                raise exc.UnauthorizedAuthError(
                    f"No invitation found for token {cmd.token}"
                )
            # Choose the invitation with the latest expiry date
            user_invitation = sorted(
                user_invitations_with_token, key=lambda x: x.expires_at
            )[-1]
            # Set roles of the user
            new_user.roles = user_invitation.roles
            # Set ID and organization ID of the user
            new_user.organization_id = user_invitation.organization_id
            # Create user
            user_in_db: model.User = self.app.user_manager.create_new_user_from_token(  # type: ignore[assignment]
                new_user,
                user_invitation.token,
                created_by_user_id=user_invitation.invited_by_user_id,
                roles=user_invitation.roles,
            )
            # Delete invitation
            self.repository.crud(
                uow,
                None,
                self.user_invitation_class,
                None,
                [x.id for x in user_invitations],
                CrudOperation.DELETE_SOME,
            )
        return user_in_db

    def update_user(
        self,
        cmd: command.UpdateUserCommand,
    ) -> model.User:
        assert cmd.user
        if cmd.roles is not None and len(cmd.roles) == 0:
            raise exc.InvalidArgumentsError("Roles cannot be empty")
        with self.repository.uow() as uow:
            tgt_user: model.User = self.repository.crud(  # type: ignore[assignment]
                uow,
                cmd.user.id,
                self.user_class,
                None,
                cmd.tgt_user_id,
                CrudOperation.READ_ONE,
            )
            is_active = tgt_user.is_active if cmd.is_active is None else cmd.is_active
            roles = tgt_user.roles if cmd.roles is None else cmd.roles
            organization_id = (
                tgt_user.organization_id
                if cmd.organization_id is None
                else cmd.organization_id
            )
            # Special case: no updates
            if (
                tgt_user.is_active == is_active
                and tgt_user.roles == roles
                and tgt_user.organization_id == organization_id
            ):
                return tgt_user
            # Check if organization_id exists
            if tgt_user.organization_id != organization_id:
                self.repository.crud(
                    uow,
                    cmd.user.id,
                    model.Organization,
                    None,
                    organization_id,
                    CrudOperation.READ_ONE,
                )
            # Update user
            tgt_user.is_active = is_active
            tgt_user.roles = roles
            tgt_user.organization_id = organization_id
            updated_tgt_user: model.User = self.repository.crud(  # type: ignore[assignment]
                uow,
                cmd.user.id,
                self.user_class,
                tgt_user,
                None,
                CrudOperation.UPDATE_ONE,
            )

        # Invalidate cache for the user
        self.retrieve_user_by_key.cache_clear()  # type: ignore[attr-defined]
        return updated_tgt_user

    def retrieve_organization_contact(
        self,
        cmd: command.RetrieveOrganizationContactCommand,
    ) -> list[model.Contact]:
        user, repository = self._get_user_and_repository(cmd)

        sites: list[model.Site]
        contacts: list[model.Contact]
        with repository.uow() as uow:
            if cmd.organization_ids:
                sites = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Site,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                )
                contacts = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Contact,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                )
                organization_ids = set(cmd.organization_ids)
                sites = [x for x in sites if x.organization_id in organization_ids]
                site_ids = {x.id for x in sites}
                contacts = [x for x in contacts if x.site_id in site_ids]
            elif cmd.site_ids:
                sites = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Site,
                    None,
                    cmd.site_ids,
                    CrudOperation.READ_SOME,
                )
                contacts = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Contact,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                )
                site_ids = {x.id for x in sites}
                contacts = [x for x in contacts if x.site_id in site_ids]
            elif cmd.contact_ids:
                contacts = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Contact,
                    None,
                    cmd.contact_ids,
                    CrudOperation.READ_SOME,
                )
            else:
                raise AssertionError

        return contacts



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


import json
from typing import ClassVar
from uuid import UUID

from pydantic import Field, field_serializer, field_validator

from gen_epix.commondb.domain.model import Model
from gen_epix.fastapp.domain import Entity, create_keys, create_links
from gen_epix.seqdb.domain import enum
from gen_epix.seqdb.domain.model.seq.base import ProtocolMixin, SeqMixin


class TreeAlgorithmClass(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="tree_algorithm_classes",
        table_name="tree_algorithm_class",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(
        description="The code of the tree algorithm class", max_length=255
    )
    name: str = Field(
        description="The name of the tree algorithm class", max_length=255
    )
    is_seq_based: bool = Field(
        description="Whether the sequence or alignment is needed as input"
    )
    is_dist_based: bool = Field(
        description="Whether the distance between sequences is needed as input"
    )
    rank: int | None = Field(
        default=None,
        description="The rank of the tree algorithm class, if relevant.",
    )


class TreeAlgorithm(Model):
    """
    See https://en.wikipedia.org/wiki/Hierarchical_clustering,
    https://en.wikipedia.org/wiki/Neighbor_joining,
     https://en.wikipedia.org/wiki/Computational_phylogenetics,
     https://en.wikipedia.org/wiki/Spanning_tree
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="tree_algorithms",
        table_name="tree_algorithm",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
        links=create_links(
            {
                1: (
                    "tree_algorithm_class_id",
                    TreeAlgorithmClass,
                    "tree_algorithm_class",
                ),
            }
        ),
    )
    code: enum.TreeAlgorithm = Field(description="The code of the tree algorithm")
    name: str = Field(description="The name of the tree algorithm", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the tree algorithm"
    )
    tree_algorithm_class_id: UUID = Field(
        description="The ID of the tree algorithm class. FOREIGN KEY"
    )
    tree_algorithm_class: TreeAlgorithmClass | None = Field(
        default=None, description="The class of algorithm"
    )
    is_ultrametric: bool = Field(description="Whether the tree is ultrametric")
    rank: int | None = Field(
        default=None,
        description="The rank of the tree algorithm, if relevant.",
    )


class LibraryPrepProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="library_prep_protocols",
        table_name="library_prep_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )


class AssemblyProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="assembly_protocols",
        table_name="assembly_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )
    has_manual_curation: bool = Field(
        default=False,
        description="Whether the assembly has a, potentially optional, manual curation step.",
    )


class LocusDetectionProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="locus_detection_protocols",
        table_name="locus_detection_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )


class SnpDetectionProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="snp_detection_protocols",
        table_name="snp_detection_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )


class KmerDetectionProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="kmer_detection_protocols",
        table_name="kmer_detection_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )


class AlignmentProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="alignment_protocols",
        table_name="alignment_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )
    is_multiple: bool = Field(
        description="Whether the alignment protocol can be used for more than two sequences"
    )


class TaxonomyProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="taxonomy_protocols",
        table_name="taxonomy_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )


class SeqClassificationProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_classification_protocols",
        table_name="seq_classification_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )
    is_taxonomic: bool = Field(
        description="Whether the category is based on phylogeny or not"
    )


class PcrProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="pcr_protocols",
        table_name="pcr_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )
    target_names: list[str] = Field(
        description="The names of the targets for which PCR is performed"
    )

    @field_validator("target_names", mode="before")
    def _validate_target_names(cls, value: list[str] | str) -> list[str]:
        if isinstance(value, str):
            return json.loads(value)
        return value


class AstProtocol(Model, ProtocolMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ast_protocols",
        table_name="ast_protocol",
        persistable=True,
        keys=create_keys({1: "code", 2: ("name", "version")}),
    )
    is_predicted: bool = Field(
        description="Whether the AST results are predicted from sequence data"
    )
    antimicrobial_names: list[str] = Field(
        description="The names of the antimicrobials for which AST is performed"
    )

    @field_validator("antimicrobial_names", mode="before")
    def _validate_antimicrobial_names(cls, value: list[str] | str) -> list[str]:
        if isinstance(value, str):
            return json.loads(value)
        return value


class SeqCategorySet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_category_sets",
        table_name="seq_category_set",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(description="The code of the category set", max_length=255)
    name: str = Field(description="The name of the category set", max_length=255)


class SeqCategory(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="seq_categories",
        table_name="seq_category",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
        links=create_links(
            {
                1: (
                    "seq_category_set_id",
                    SeqCategorySet,
                    "seq_category_set",
                )
            }
        ),
    )
    code: str = Field(description="The code of the category", max_length=255)
    name: str = Field(description="The name of the category", max_length=255)
    seq_category_set_id: UUID = Field(
        description="The ID of the sequence category set. FOREIGN KEY"
    )
    seq_category_set: SeqCategorySet = Field(description="The sequence category set")


class SubtypingScheme(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="subtyping_schemes",
        table_name="subtyping_scheme",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(description="The code of the subtyping scheme", max_length=255)
    name: str = Field(description="The name of the subtyping scheme", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the subtyping scheme"
    )


class Taxon(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="taxa",
        table_name="taxon",
        persistable=True,
        keys=create_keys({1: "code"}),
        links=create_links(
            {
                1: ("subtyping_scheme_id", SubtypingScheme, "subtyping_scheme"),
            }
        ),
    )
    code: str = Field(description="The code of the taxon", max_length=255)
    name: str = Field(description="The name of the taxon", max_length=255)
    rank: enum.TaxonRank = Field(description="The rank of the taxon")
    ncbi_taxid: int | None = Field(
        default=None, description="The NCBI Taxonomy ID of the taxon"
    )
    ictv_ictv_id: str | None = Field(
        default=None, description="The ICTV ID of the taxon", max_length=255
    )
    snomed_sctid: int | None = Field(
        default=None, description="The Snomed CT ID of the taxon"
    )
    subtyping_scheme_id: UUID | None = Field(
        default=None, description="The ID of the subtyping scheme, if any. FOREIGN KEY"
    )
    subtyping_scheme: SubtypingScheme | None = Field(
        default=None, description="The subtyping scheme"
    )
    ncbi_ancestor_taxids: list[int] | None = Field(
        default=None,
        description="The NCBI taxon IDs of the ancestors, sorted from highest to lowest rank",
    )
    ancestor_taxon_ids: list[UUID] = Field(
        description="The IDs of the ancestor taxa, sorted from highest to lowest rank"
    )

    @field_validator("ncbi_ancestor_taxids", mode="before")
    @classmethod
    def _validate_ncbi_ancestor_taxids(cls, value: list[int] | str) -> list[int]:
        if isinstance(value, str):
            return [int(x) for x in json.loads(value)]
        return value

    @field_validator("ancestor_taxon_ids", mode="before")
    @classmethod
    def _validate_ancestor_taxon_ids(cls, value: list[UUID] | str) -> list[UUID]:
        if isinstance(value, str):
            return [UUID(x) for x in json.loads(value)]
        return value

    @field_serializer("ancestor_taxon_ids", mode="plain")
    def _serialize_ancestor_taxon_ids(self, value: list[UUID]) -> list[str]:
        return [str(x) for x in value]


class TaxonSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="taxon_sets",
        table_name="taxon_set",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(description="The code of the taxon set", max_length=255)
    name: str = Field(description="The name of the taxon set", max_length=255)


class TaxonSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="taxon_set_members",
        table_name="taxon_set_member",
        persistable=True,
        keys=create_keys({1: "taxon_set_id", 2: "taxon_id"}),
        links=create_links(
            {
                1: ("taxon_set_id", TaxonSet, "taxon_set"),
                2: ("taxon_id", Taxon, "taxon"),
            }
        ),
    )
    taxon_set_id: UUID = Field(description="The ID of the taxon set. FOREIGN KEY")
    taxon_set: TaxonSet = Field(description="The taxon set")
    taxon_id: UUID = Field(description="The ID of the taxon. FOREIGN KEY")
    taxon: Taxon = Field(description="The taxon")


class Locus(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="loci",
        table_name="locus",
        persistable=True,
        keys=create_keys({1: "code"}),
    )
    code: str = Field(description="The code of the locus.", max_length=255)
    gene_code: str | None = Field(
        default=None,
        description="The code of the gene, if the locus corresponds to one and a code is available.",
        max_length=255,
    )
    product_name: str | None = Field(
        default=None,
        description="The name of the gene product, if available.",
        max_length=255,
    )


class LocusSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="locus_sets",
        table_name="locus_set",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(description="The code of the locus set.", max_length=255)
    name: str = Field(description="The name of the locus set.", max_length=255)
    n_loci: int = Field(description="The number of loci in the locus set.")
    locus_ids: list[UUID] = Field(
        description="The ordered IDs of the loci in the locus set."
    )

    @field_validator("locus_ids", mode="before")
    @classmethod
    def _validate_locus_ids(cls, value: list[UUID] | str) -> list[UUID]:
        """
        Validate and convert locus_ids representation to a list[UUID]. When given as a
        string, it is assumed to be a JSON list of UUID string representations.
        """
        if isinstance(value, str):
            return [UUID(x) for x in json.loads(value)]
        return value

    @field_serializer("locus_ids", mode="plain")
    def _serialize_locus_ids(self, value: list[UUID]) -> list[str]:
        return [str(x) for x in value]


class LocusSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="locus_set_members",
        table_name="locus_set_member",
        persistable=True,
        keys=create_keys({1: ("locus_set_id", "locus_id")}),
        links=create_links(
            {
                1: ("locus_set_id", LocusSet, "locus_set"),
                2: ("locus_id", Locus, "locus"),
            }
        ),
    )
    locus_set_id: UUID = Field(
        description="The unique identifier for the locus set. FOREIGN KEY"
    )
    locus_set: LocusSet | None = Field(default=None, description="The locus set.")
    locus_id: UUID = Field(
        description="The unique identifier for the locus. FOREIGN KEY"
    )
    locus: Locus | None = Field(default=None, description="The locus.")
    index: int = Field(
        description="The index (ordinal number) of the locus in the locus set."
    )


class TaxonLocusLink(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="taxon_locus_links",
        table_name="taxon_locus_link",
        persistable=True,
        keys=create_keys({1: ("taxon_id", "locus_id")}),
        links=create_links(
            {1: ("taxon_id", Taxon, "taxon"), 2: ("locus_id", Locus, "locus")}
        ),
    )
    taxon_id: UUID = Field(
        description="The unique identifier for the taxon. FOREIGN KEY"
    )
    taxon: Taxon | None = Field(default=None, description="The taxon.")
    locus_id: UUID = Field(
        description="The unique identifier for the locus. FOREIGN KEY"
    )
    locus: Locus | None = Field(default=None, description="The locus.")


class RefSeq(Model, SeqMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ref_seqs",
        table_name="ref_seq",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
        links=create_links({1: ("taxon_id", Taxon, "taxon")}),
    )
    code: str = Field(description="The code of the reference sequence", max_length=255)
    name: str = Field(description="The name of the reference sequence", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the reference sequence"
    )
    taxon_id: UUID = Field(description="The ID of the taxon. FOREIGN KEY")
    taxon: Taxon | None = Field(default=None, description="The taxon")
    genbank_accession_code: str | None = Field(
        default=None,
        description="The GenBank accession code of the reference sequence",
        max_length=255,
    )


class RefAllele(Model, SeqMixin):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ref_alleles",
        table_name="ref_allele",
        persistable=True,
        keys=create_keys({1: ("locus_id", "index")}),
        links=create_links({1: ("locus_id", Locus, "locus")}),
    )
    locus_id: UUID = Field(
        description="The unique identifier for the locus. FOREIGN KEY"
    )
    locus: Locus | None = Field(default=None, description="The locus.")
    index: int = Field(
        description="The index (ordinal number) of the reference allele for the locus."
    )


class RefSnp(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ref_snps",
        table_name="ref_snp",
        persistable=True,
        keys=create_keys({1: "code", 2: ("ref_seq_id", "position", "nucleotide")}),
        links=create_links({1: ("ref_seq_id", RefSeq, "ref_seq")}),
    )
    code: str = Field(description="The code of the reference SNP.", max_length=255)
    ref_seq_id: UUID = Field(
        description="The unique identifier for the reference sequence. FOREIGN KEY"
    )
    ref_seq: RefSeq | None = Field(default=None, description="The reference sequence.")
    position: int = Field(description="The position of the reference SNP.")
    nucleotide: str = Field(
        description="The nucleotide of the reference SNP.", min_length=1, max_length=1
    )


class RefSnpSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ref_snp_sets",
        table_name="ref_snp_set",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(description="The code of the reference SNP set.", max_length=255)
    name: str = Field(description="The name of the reference SNP set.", max_length=255)


class RefSnpSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ref_snp_set_members",
        table_name="ref_snp_set_member",
        persistable=True,
        keys=create_keys(
            {
                1: ("ref_snp_set_id", "ref_snp_id"),
                2: ("ref_snp_set_id", "index"),
            }
        ),
        links=create_links(
            {
                1: ("ref_snp_set_id", RefSnpSet, "ref_snp_set"),
                2: ("ref_snp_id", RefSnp, "ref_snp"),
            }
        ),
    )
    ref_snp_set_id: UUID = Field(
        description="The unique identifier for the reference SNP set. FOREIGN KEY"
    )
    ref_snp_set: RefSnpSet | None = Field(
        default=None, description="The reference SNP set."
    )
    ref_snp_id: UUID = Field(
        description="The unique identifier for the reference SNP. FOREIGN KEY"
    )
    ref_snp: RefSnp | None = Field(default=None, description="The reference SNP.")
    index: int = Field(
        description="The index (ordinal number) of the reference SNP in the reference SNP set."
    )



from typing import ClassVar

import gen_epix.commondb.domain.model.system as model
from gen_epix.commondb.domain.command.base import Command, CrudCommand

# Non-CRUD commands


class RetrieveOutagesCommand(Command):
    pass


class RetrieveLicensesCommand(Command):
    pass


# CRUD commands


class OutageCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Outage



"""
Transformer registry for managing and creating transformer instances.
"""

from collections.abc import Callable
from typing import Any, TypeVar

from gen_epix.transform.transformer import Transformer

T = TypeVar("T", bound=Transformer)


class Registry:
    """Central registry for transformer types and factory methods."""

    _transformers: dict[str, type[Transformer]] = {}
    _factories: dict[str, Callable[..., Transformer]] = {}

    @classmethod
    def register(cls, name: str, transformer_class: type[Transformer]) -> None:
        """Register a transformer class by name."""
        cls._transformers[name] = transformer_class

    @classmethod
    def register_factory(
        cls, name: str, factory_fn: Callable[..., Transformer]
    ) -> None:
        """Register a factory function for creating transformer instances."""
        cls._factories[name] = factory_fn

    @classmethod
    def create(cls, name: str, **kwargs: Any) -> Transformer:
        """Create transformer instance by name."""
        if name in cls._factories:
            return cls._factories[name](**kwargs)
        elif name in cls._transformers:
            return cls._transformers[name](**kwargs)
        else:
            raise ValueError(f"Unknown transformer: {name}")

    @classmethod
    def list_available(cls) -> list[str]:
        """List all available transformer names."""
        return list(set(cls._transformers.keys()) | set(cls._factories.keys()))

    @classmethod
    def decorator(cls, name: str) -> Callable[[type[T]], type[T]]:
        """Decorator for registering transformer classes."""

        def wrapper(transformer_class: type[T]) -> type[T]:
            cls.register(name, transformer_class)
            return transformer_class

        return wrapper

    @classmethod
    def factory_decorator(
        cls, name: str
    ) -> Callable[[Callable[..., T]], Callable[..., T]]:
        """Decorator for registering factory functions."""

        def wrapper(factory_fn: Callable[..., T]) -> Callable[..., T]:
            cls.register_factory(name, factory_fn)
            return factory_fn

        return wrapper

    @classmethod
    def clear(cls) -> None:
        """Clear all registered transformers and factories."""
        cls._transformers.clear()
        cls._factories.clear()


# Convenience decorators
def register_transformer(name: str) -> Callable[[type[T]], type[T]]:
    """Decorator to register a transformer class."""
    return Registry.decorator(name)


def register_factory(name: str) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Decorator to register a transformer factory function."""
    return Registry.factory_decorator(name)



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


import json
from collections.abc import Iterable
from datetime import datetime
from typing import Any, ClassVar, Self
from uuid import UUID

from pydantic import Field, field_serializer, field_validator, model_validator

from gen_epix.casedb.domain import enum, exc
from gen_epix.casedb.domain.model.geo import RegionSet
from gen_epix.casedb.domain.model.ontology import ConceptSet, Disease, EtiologicalAgent
from gen_epix.casedb.domain.model.subject import Subject
from gen_epix.commondb.domain.model import DataCollection, Model
from gen_epix.fastapp.domain import Entity, create_keys, create_links


class GeneticDistanceProtocol(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="genetic_distance_protocols",
        table_name="genetic_distance_protocol",
        persistable=True,
        keys=create_keys({1: "seqdb_seq_distance_protocol_id", 2: "name"}),
    )
    seqdb_seq_distance_protocol_id: UUID = Field(
        description="The ID of the protocol in seqdb"
    )
    name: str = Field(description="The name of the protocol", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the protocol"
    )
    seqdb_max_stored_distance: float | None = Field(
        default=None,
        description="The maximum distance that is stored in seqdb for this protocol",
    )
    min_scale_unit: float = Field(description="The minimum unit to be shown in a scale")


class TreeAlgorithmClass(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="tree_algorithm_classes",
        table_name="tree_algorithm_class",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(
        description="The code of the tree algorithm class", max_length=255
    )
    name: str = Field(
        description="The name of the tree algorithm class", max_length=255
    )
    is_seq_based: bool = Field(
        description="Whether the sequence or alignment is needed as input"
    )
    is_dist_based: bool = Field(
        description="Whether the distance between sequences is needed as input"
    )
    rank: int | None = Field(
        default=None,
        description="The rank of the tree algorithm class, if relevant.",
    )


class TreeAlgorithm(Model):
    """
    See https://en.wikipedia.org/wiki/Hierarchical_clustering,
    https://en.wikipedia.org/wiki/Neighbor_joining,
     https://en.wikipedia.org/wiki/Computational_phylogenetics,
     https://en.wikipedia.org/wiki/Spanning_tree
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="tree_algorithms",
        table_name="tree_algorithm",
        persistable=True,
        keys=create_keys({1: "seqdb_tree_algorithm_id", 2: "code", 3: "name"}),
        links=create_links(
            {
                1: (
                    "tree_algorithm_class_id",
                    TreeAlgorithmClass,
                    "tree_algorithm_class",
                ),
            }
        ),
    )
    tree_algorithm_class_id: UUID = Field(
        description="The ID of the tree algorithm class. FOREIGN KEY"
    )
    tree_algorithm_class: TreeAlgorithmClass | None = Field(
        default=None, description="The class of algorithm"
    )
    seqdb_tree_algorithm_id: UUID = Field(
        description="The ID of the tree algorithm in seqdb"
    )
    code: enum.TreeAlgorithmType = Field(description="The code of the tree algorithm")
    name: str = Field(description="The name of the tree algorithm", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the tree algorithm"
    )
    is_ultrametric: bool = Field(description="Whether the tree is ultrametric")
    rank: int | None = Field(
        default=None,
        description="The rank of the tree algorithm, if relevant.",
    )


class Dim(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="dims",
        table_name="dim",
        persistable=True,
        keys=create_keys({1: "code"}),
    )
    dim_type: enum.DimType = Field(description="The type of dimension.")
    code: str = Field(description="The code for the dimension.", max_length=255)
    label: str = Field(description="The label for the dimension.")
    rank: int | None = Field(
        default=None,
        description="The rank of the dimension, if relevant.",
    )
    col_code_prefix: str | None = Field(
        default=None,
        description=(
            "The column code prefix used to compose a full column code,"
            " if different from the code field."
        ),
    )
    description: str | None = Field(
        default=None, description="Description of the dimension."
    )
    props: dict[str, Any] = Field(
        default_factory=dict, description="Additional properties of the dimension."
    )

    @field_validator("code", mode="before")
    @classmethod
    def validate_code(cls, value: Any) -> str:
        return str(value)


class Col(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="cols",
        table_name="col",
        persistable=True,
        keys=create_keys({1: ("dim_id", "code")}),
        links=create_links(
            {
                1: ("dim_id", Dim, "dim"),
                2: ("concept_set_id", ConceptSet, "concept_set"),
                3: ("region_set_id", RegionSet, "region_set"),
                4: (
                    "genetic_distance_protocol_id",
                    GeneticDistanceProtocol,
                    "genetic_distance_protocol",
                ),
            }
        ),
    )
    dim_id: UUID = Field(description="The ID of the dimension. FOREIGN KEY")
    dim: Dim | None = Field(default=None, description="The dimension")
    code_suffix: str | None = Field(
        default=None,
        description=(
            "The code suffix for the column used to compose a full column code,"
            " if needed in addition to the dimension column code prefix. See code field."
        ),
    )
    code: str = Field(
        description=(
            "The code for the column, "
            "equal to the dimension column code prefix dot code_suffix"
            " (dot code_suffix only if the latter is not null)."
        ),
        max_length=255,
    )
    rank_in_dim: int | None = Field(
        default=None,
        description="The rank of the column within the dimension, if relevant.",
    )
    label: str | None = Field(
        default=None,
        description="The label for the column, if different from the code.",
    )
    col_type: enum.ColType = Field(
        description="The type of the data stored in the column."
    )
    concept_set_id: UUID | None = Field(
        default=None,
        description=(
            "The ID of the concept set for the column in case of type"
            " NOMINAL, ORDINAL, INTERVAL. FOREIGN KEY"
        ),
    )
    concept_set: ConceptSet | None = Field(default=None, description="The concept set.")
    region_set_id: UUID | None = Field(
        default=None,
        description="The ID of the region set for the column in case of type GEO. FOREIGN KEY",
    )
    region_set: RegionSet | None = Field(default=None, description="The region set.")
    genetic_distance_protocol_id: UUID | None = Field(
        default=None,
        description=(
            "The ID of the genetic distance protocol"
            " that produces the input for the tree algorithm. FOREIGN KEY"
        ),
    )
    genetic_distance_protocol: GeneticDistanceProtocol | None = Field(
        default=None, description="The genetic distance protocol"
    )
    description: str | None = Field(
        default=None, description="Description of the column."
    )
    props: dict[str, Any] = Field(
        default_factory=dict, description="Additional properties of the column."
    )

    @field_validator("code", mode="before")
    @classmethod
    def validate_code(cls, value: Any) -> str:
        return str(value)

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if self.col_type in enum.ColTypeSet.HAS_CONCEPT_SET.value:
            if self.concept_set_id is None:
                raise exc.InvalidArgumentsError(
                    f"No concept_set_id provided for col_type {self.col_type.value}"
                )
        if self.col_type in enum.ColTypeSet.HAS_REGION_SET.value:
            if self.region_set_id is None:
                raise exc.InvalidArgumentsError(
                    f"No region_set_id provided for col_type {self.col_type.value}"
                )
        if self.col_type in enum.ColTypeSet.HAS_GENETIC_DISTANCE_PROTOCOL.value:
            if self.genetic_distance_protocol_id is None:
                raise exc.InvalidArgumentsError(
                    f"No genetic_distance_protocol_id provided for col_type {self.col_type.value}"
                )
        return self


class CaseType(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_types",
        table_name="case_type",
        persistable=True,
        keys=create_keys({1: "name"}),
        links=create_links(
            {
                1: ("disease_id", Disease, "disease"),
                2: ("etiological_agent_id", EtiologicalAgent, "etiological_agent"),
            }
        ),
    )
    name: str = Field(description="The name of the case type", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the case type"
    )
    disease_id: UUID | None = Field(
        default=None, description="The ID of the disease. FOREIGN KEY"
    )
    disease: Disease | None = Field(default=None, description="The disease")
    etiological_agent_id: UUID | None = Field(
        default=None, description="The ID of the etiological agent. FOREIGN KEY"
    )
    etiological_agent: EtiologicalAgent | None = Field(
        default=None, description="The etiological agent"
    )


class CaseTypeSetCategory(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_set_categories",
        table_name="case_type_set_category",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    name: str = Field(
        description="The name of the case type set category", max_length=255
    )
    description: str | None = Field(
        default=None, description="The description of the case type set category"
    )
    rank: int = Field(description="The rank of the case type set category")
    purpose: enum.CaseTypeSetCategoryPurpose = Field(
        default=enum.CaseTypeSetCategoryPurpose.CONTENT,
        description="The purpose of the case type set category",
    )

    @field_serializer("purpose", mode="plain")
    def _serialize_purpose(self, value: enum.CaseTypeSetCategoryPurpose) -> str:
        return value.value


class CaseTypeSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_sets",
        table_name="case_type_set",
        persistable=True,
        keys=create_keys({1: ("case_type_set_category_id", "name")}),
        links=create_links(
            {
                1: (
                    "case_type_set_category_id",
                    CaseTypeSetCategory,
                    "case_type_set_category",
                )
            }
        ),
    )
    name: str = Field(description="The name of the case type set", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the case type set"
    )
    case_type_set_category_id: UUID = Field(
        description="The id of the category of the case type set. FOREIGN KEY"
    )
    case_type_set_category: CaseTypeSetCategory | None = Field(
        default=None, description="The category of the case type set"
    )
    rank: float = Field(
        description="The rank of the case type set, establishing a partial order"
    )


class CaseTypeSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_set_members",
        table_name="case_type_set_member",
        persistable=True,
        keys=create_keys({1: ("case_type_set_id", "case_type_id")}),
        links=create_links(
            {
                1: ("case_type_set_id", CaseTypeSet, "case_type_set"),
                2: ("case_type_id", CaseType, "case_type"),
            }
        ),
    )
    case_type_set_id: UUID = Field(
        description="The ID of the case type set. FOREIGN KEY"
    )
    case_type_set: CaseTypeSet | None = Field(
        default=None, description="The case type set"
    )
    case_type_id: UUID = Field(description="The ID of the case type. FOREIGN KEY")
    case_type: CaseType | None = Field(default=None, description="The case type")


class CaseTypeCol(Model):  # type: ignore
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_cols",
        table_name="case_type_col",
        persistable=True,
        keys=create_keys({1: ("case_type_id", "col_id", "occurrence")}),
        links=create_links(
            {
                1: ("case_type_id", CaseType, "case_type"),
                2: ("col_id", Col, "col"),
            }
        ),
    )
    case_type_id: UUID = Field(description="The ID of the case type. FOREIGN KEY")
    case_type: CaseType | None = Field(default=None, description="The case type")
    col_id: UUID = Field(description="The ID of the column. FOREIGN KEY")
    col: Col | None = Field(default=None, description="The column")
    occurrence: int | None = Field(
        default=None,
        description=(
            "The index of the occurrence of the column for this case type."
            " E.g. for first and second vaccination date it would be 1 and 2."
            " Empty or 1 if only a single occurrence."
        ),
    )
    code: str = Field(
        description=(
            "The code for the case type column, "
            "equal to the column code and, if present, dot 'x' occurrence. "
            "E.g. 'Host.Vaccination.Date.COVID19.x1' for occurrence=1, "
            "'Specimen.Sampling.Date' for occurrence null"
        ),
        max_length=255,
    )
    rank: int | None = Field(
        default=None,
        description=(
            "The rank of the column for this case type for ordering, "
            "if different from the general dimension and column rank."
        ),
    )
    label: str | None = Field(
        default=None,
        description=(
            "The label of the column for this case type,"
            " if different from the general column label."
        ),
    )
    description: str | None = Field(
        default=None, description="Description of the case type column."
    )
    min_value: float | None = Field(
        default=None, description="The minimum value for a numeric column"
    )
    max_value: float | None = Field(
        default=None, description="The maximum value for a numeric column"
    )
    min_datetime: datetime | None = Field(
        default=None, description="The minimum datetime for a time column"
    )
    max_datetime: datetime | None = Field(
        default=None, description="The maximum datetime for a time column"
    )
    min_length: int | None = Field(
        default=None, description="The minimum length for a text column, if not empty"
    )
    max_length: int | None = Field(
        default=None, description="The maximum length for a text column, if not empty"
    )
    pattern: str | None = Field(
        default=None,
        description="The regular expression for a text column, if not empty",
    )
    ncbi_taxid: str | None = Field(
        default=None,
        description=(
            "The NCBI taxid for the column, if the column is a genetic sequence"
        ),
        pattern=r"^NCBI:txid\d+$",
    )
    genetic_sequence_case_type_col_id: UUID | None = Field(
        default=None,
        description=(
            "The ID of the genetic sequence case type column, "
            "if this is a genetic sequence column. FOREIGN KEY"
        ),
    )
    tree_algorithm_codes: set[enum.TreeAlgorithmType] | None = Field(
        default=None,
        description=(
            "The set of tree algorithms that can be used for the case type column"
        ),
    )
    props: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional properties of the case type column.",
    )

    @field_validator("code", mode="before")
    @classmethod
    def validate_code(cls, value: Any) -> str:
        return str(value)

    @field_validator("tree_algorithm_codes", mode="before")
    @classmethod
    def validate_tree_algorithm_codes(
        cls, value: Iterable[enum.TreeAlgorithmType] | str | None
    ) -> set[enum.TreeAlgorithmType] | None:
        if value is None or isinstance(value, set):
            return value
        if isinstance(value, str):
            return {enum.TreeAlgorithmType[x] for x in json.loads(value)}
        return set(value)

    @field_serializer("tree_algorithm_codes", mode="plain")
    def _serialize_tree_algorithm_codes(
        self, value: list[enum.TreeAlgorithmType] | None
    ) -> list[str] | None:
        return None if value is None else [x.value for x in value]


class CaseTypeColSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_col_sets",
        table_name="case_type_col_set",
        persistable=True,
        keys=create_keys({1: "name"}),
        # links=get_links({
        #     1: ("case_type_id", CaseType, "case_type"),
        # }),
    )
    # case_type_id: UUID = Field(description="The ID of the case type. FOREIGN KEY")
    # case_type: CaseType | None = Field(default=None, description="The case type")
    name: str = Field(
        description="The name of a case type column set, UNIQUE", max_length=255
    )
    description: str | None = Field(
        default=None, description="The description of the case type column set"
    )


class CaseTypeColSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_col_set_members",
        table_name="case_type_col_set_member",
        persistable=True,
        keys=create_keys({1: ("case_type_col_set_id", "case_type_col_id")}),
        links=create_links(
            {
                1: ("case_type_col_set_id", CaseTypeColSet, "case_type_col_set"),
                2: ("case_type_col_id", CaseTypeCol, "case_type_col"),  # type: ignore
            }
        ),
    )
    case_type_col_set_id: UUID = Field(
        description="The ID of the case type column set. FOREIGN KEY"
    )
    case_type_col_set: CaseTypeColSet | None = Field(
        default=None, description="The case type column set"
    )
    case_type_col_id: UUID = Field(
        description="The ID of the case type column. FOREIGN KEY"
    )
    case_type_col: CaseTypeCol | None = Field(  # type: ignore
        default=None, description="The case type column"
    )


class Case(Model):
    """
    A class representing a case.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="cases",
        table_name="case",
        persistable=True,
        links=create_links(
            {
                1: ("case_type_id", CaseType, "case_type"),
                2: ("subject_id", Subject, "subject"),
                3: (
                    "created_in_data_collection_id",
                    DataCollection,
                    "created_in_data_collection",
                ),
            }
        ),
    )
    case_type_id: UUID = Field(description="The ID of the case type. FOREIGN KEY")
    case_type: CaseType | None = Field(default=None, description="The case type")
    subject_id: UUID | None = Field(
        default=None, description="The ID of the subject. FOREIGN KEY"
    )
    subject: Subject | None = Field(default=None, description="The subject")
    created_in_data_collection_id: UUID = Field(
        description="The ID of the data collection where the case was created. FOREIGN KEY",
    )
    created_in_data_collection: DataCollection | None = Field(
        default=None, description="The data collection where the case was created"
    )
    count: int | None = Field(
        default=None, description="The number of cases, if applicable", gt=0
    )
    case_date: datetime = Field(description="The date of the case")
    content: dict[UUID, str] = Field(
        description="The column data of the case as {col_id: str_value}"
    )

    @field_serializer("content", mode="plain")
    def _serialize_content(self, value: dict[UUID, str]) -> dict[str, str]:
        return {str(x): y for x, y in value.items()}


class CaseDataCollectionLink(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_data_collection_links",
        table_name="case_data_collection_link",
        persistable=True,
        keys=create_keys({1: ("case_id", "data_collection_id")}),
        links=create_links(
            {
                1: ("case_id", Case, "case"),
                2: ("data_collection_id", DataCollection, "data_collection"),
            }
        ),
    )
    case_id: UUID = Field(description="The ID of the case. FOREIGN KEY")
    case: Case | None = Field(default=None, description="The case")
    data_collection_id: UUID = Field(
        description="The ID of the data collection. FOREIGN KEY"
    )
    data_collection: DataCollection | None = Field(
        default=None, description="The data collection"
    )


class CaseSetCategory(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_set_categories",
        table_name="case_set_category",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    name: str = Field(
        description="The name of the case set category, UNIQUE", max_length=255
    )
    description: str | None = Field(
        description="The description of the case set category"
    )


class CaseSetStatus(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_set_statuses",
        table_name="case_set_status",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    name: str = Field(
        description="The name of the case set status, UNIQUE", max_length=255
    )
    description: str | None = Field(
        description="The description of the case set status"
    )


class CaseSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_sets",
        table_name="case_set",
        persistable=True,
        keys=create_keys({1: "name"}),
        links=create_links(
            {
                1: ("case_type_id", CaseType, "case_type"),
                2: (
                    "created_in_data_collection_id",
                    DataCollection,
                    "created_in_data_collection",
                ),
                3: ("case_set_category_id", CaseSetCategory, "case_set_category"),
                4: ("case_set_status_id", CaseSetStatus, "case_set_status"),
            }
        ),
    )
    case_type_id: UUID = Field(description="The ID of the case type. FOREIGN KEY")
    case_type: CaseType | None = Field(default=None, description="The case type")
    created_in_data_collection_id: UUID = Field(
        description="The ID of the data collection where the case set was created. FOREIGN KEY",
    )
    created_in_data_collection: DataCollection | None = Field(
        default=None, description="The data collection where the case set was created"
    )
    name: str = Field(description="The name of a case set, UNIQUE", max_length=255)
    description: str = Field(description="The description of a case set")
    created_at: datetime = Field(
        description="The datetime of the case set creation",
        default_factory=datetime.now,
    )
    case_set_category_id: UUID = Field(
        description="The id of the category of the case set"
    )
    case_set_category: CaseSetCategory | None = Field(
        default=None, description="The category of the case set"
    )
    case_set_status_id: UUID = Field(description="The id of the status of the case set")
    case_set_status: CaseSetStatus | None = Field(
        default=None, description="The status of the case set"
    )


class CaseSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_set_members",
        table_name="case_set_member",
        persistable=True,
        keys=create_keys({1: ("case_set_id", "case_id")}),
        links=create_links(
            {1: ("case_set_id", CaseSet, "case_set"), 2: ("case_id", Case, "case")}
        ),
    )
    case_set_id: UUID = Field(description="The ID of the case set. FOREIGN KEY")
    case_set: CaseSet | None = Field(default=None, description="The case set")
    case_id: UUID = Field(description="The ID of the case. FOREIGN KEY")
    case: Case | None = Field(default=None, description="The case")
    classification: enum.CaseClassification | None = Field(
        default=None, description="The classification of the case"
    )


class CaseSetDataCollectionLink(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_set_data_collection_links",
        table_name="case_set_data_collection_link",
        persistable=True,
        keys=create_keys({1: ("case_set_id", "data_collection_id")}),
        links=create_links(
            {
                1: ("case_set_id", CaseSet, "case_set"),
                2: ("data_collection_id", DataCollection, "data_collection"),
            }
        ),
    )
    case_set_id: UUID = Field(description="The ID of the case set. FOREIGN KEY")
    case_set: CaseSet | None = Field(default=None, description="The case set")
    data_collection_id: UUID = Field(
        description="The ID of the data collection. FOREIGN KEY"
    )
    data_collection: DataCollection | None = Field(
        default=None, description="The data collection"
    )



import importlib
from collections.abc import Hashable, Iterable
from typing import Any
from uuid import UUID

from gen_epix.casedb.domain import command
from gen_epix.casedb.domain import enum as enum
from gen_epix.casedb.domain import exc, model
from gen_epix.casedb.domain.service import BaseSeqdbService
from gen_epix.commondb.config import AppCfg
from gen_epix.commondb.domain.enum import AppType
from gen_epix.fastapp import App, CrudCommand, Model
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.seqdb.domain import command as seqdb_command
from gen_epix.seqdb.domain import enum as seqdb_enum
from gen_epix.seqdb.domain import model as seqdb_model
from gen_epix.seqdb.domain.model import User as SeqdbUser
from gen_epix.seqdb.env import AppComposer as SeqdbAppComposer


class SeqdbService(BaseSeqdbService):

    COMMAND_MAP: dict[type[command.Command], type[seqdb_command.Command]] = {
        command.RetrievePhylogeneticTreeBySequencesCommand: seqdb_command.RetrievePhylogeneticTreeCommand,
    }
    TREE_ALGORITHM_MAP = {
        x: y
        for x in enum.TreeAlgorithmType
        for y in seqdb_enum.TreeAlgorithm
        if x.value == y.value
    }

    def __init__(self, app: App, seqdb_app_type: str, **kwargs: Any) -> None:
        seqdb_local_app_props = kwargs.pop("seqdb_local_app", {})
        seqdb_remote_app_props = kwargs.pop("seqdb_remote_app", {})
        super().__init__(app, **kwargs)
        seqdb_app: App
        seqdb_user: SeqdbUser | None
        if seqdb_app_type.upper() == "LOCAL":
            if "app_cfg" in seqdb_local_app_props:
                seqdb_app_cfg = seqdb_local_app_props.pop("app_cfg")
            else:
                seqdb_app_cfg = AppCfg(
                    AppType.SEQDB, seqdb_enum.ServiceType, seqdb_enum.RepositoryType
                )
            log_setup = seqdb_local_app_props.get(
                "log_setup", kwargs.get("logger") is not None
            )
            seqdb_app_composer = SeqdbAppComposer(seqdb_app_cfg, log_setup=log_setup)
            seqdb_app = seqdb_app_composer.app
            seqdb_user = SeqdbUser(**seqdb_local_app_props["user"])
        elif seqdb_app_type.upper() == "REMOTE":
            remote_app_module = seqdb_remote_app_props.pop("module")
            remote_app_class_name = seqdb_remote_app_props.pop("class_name")
            remote_app_class = getattr(
                importlib.import_module(remote_app_module), remote_app_class_name
            )
            seqdb_app = remote_app_class(**seqdb_remote_app_props)
            seqdb_user = None
        else:
            raise exc.InitializationServiceError(
                f"Invalid seqdb_app_type: {seqdb_app_type}. Must be 'LOCAL' or 'REMOTE'."
            )
        self._seqdb_app = seqdb_app
        self._seqdb_user = seqdb_user

    @property
    def seqdb_app(self) -> App:
        return self._seqdb_app

    @property
    def seqdb_user(self) -> SeqdbUser | None:
        return self._seqdb_user

    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeBySequencesCommand
    ) -> model.PhylogeneticTree | None:
        user = cmd.user
        # Prepare seqdb command and calculate tree via seqdb
        leaf_id_mapper = cmd.props.get("leaf_id_mapper")
        if leaf_id_mapper:
            leaf_names = [str(leaf_id_mapper(x)) for x in cmd.sequence_ids]
        else:
            leaf_names = None
        seqdb_cmd = seqdb_command.RetrievePhylogeneticTreeCommand(
            user=self.seqdb_user,
            seq_distance_protocol_id=cmd.seqdb_seq_distance_protocol_id,
            tree_algorithm=seqdb_enum.TreeAlgorithm[cmd.tree_algorithm_code.value],
            seq_ids=cmd.sequence_ids,
            leaf_names=leaf_names,
        )
        seqdb_phylogenetic_tree: seqdb_model.PhylogeneticTree = self.seqdb_app.handle(
            seqdb_cmd
        )
        # Convert seqdb tree model to casedb model
        phylogenetic_tree = model.PhylogeneticTree(
            tree_algorithm_code=cmd.tree_algorithm_code,
            sequence_ids=seqdb_phylogenetic_tree.seq_ids,
            leaf_ids=(
                [UUID(x) for x in seqdb_phylogenetic_tree.leaf_names]
                if seqdb_phylogenetic_tree.leaf_names is not None
                else None
            ),
            newick_repr=seqdb_phylogenetic_tree.newick_repr,
        )
        return phylogenetic_tree

    def _retrieve_seq_objects_by_ids(
        self, seq_ids: list[UUID]
    ) -> list[seqdb_model.Seq]:
        seqs: list[seqdb_model.Seq] = self.seqdb_app.handle(
            seqdb_command.SeqCrudCommand(
                user=self.seqdb_user,
                obj_ids=seq_ids,
                operation=CrudOperation.READ_SOME,
            )
        )
        return seqs

    def _retrieve_rawseq_objects_by_ids(
        self, raw_seq_ids: list[UUID]
    ) -> list[seqdb_model.RawSeq]:
        raw_seqs: list[seqdb_model.RawSeq] = self.seqdb_app.handle(
            seqdb_command.RawSeqCrudCommand(
                user=self.seqdb_user,
                obj_ids=raw_seq_ids,
                operation=CrudOperation.READ_SOME,
            )
        )
        return raw_seqs

    def retrieve_genetic_sequences(
        self, cmd: command.RetrieveGeneticSequenceByIdCommand
    ) -> list[model.GeneticSequence]:
        # naive implementation that retrieves sequences by ID
        seqs: list[seqdb_model.Seq] = self._retrieve_seq_objects_by_ids(cmd.seq_ids)
        raw_seq_ids = [seq.raw_seq_id for seq in seqs]
        raw_seqs: list[seqdb_model.RawSeq] = self._retrieve_rawseq_objects_by_ids(
            list(set(raw_seq_ids))  # type: ignore[arg-type]
        )
        raw_seq_map = {x.id: x for x in raw_seqs}
        # Convert raw sequences to model.GeneticSequence
        genetic_sequences = [
            model.GeneticSequence(
                id=seq.id, nucleotide_sequence=raw_seq_map[raw_seq_id].seq, distances={}
            )
            for seq, raw_seq_id in zip(seqs, raw_seq_ids)
        ]
        return genetic_sequences

    def retrieve_genetic_sequence_fasta_by_id(
        self,
        cmd: command.RetrieveGeneticSequenceFastaByIdCommand,
    ) -> Iterable[str]:
        seqdb_cmd = seqdb_command.RetrieveSeqFastaCommand(
            user=self.seqdb_user,
            seq_ids=cmd.seq_ids,
            wrap=cmd.wrap,
        )
        fasta_iterator: Iterable[str] = self.seqdb_app.handle(seqdb_cmd)
        return fasta_iterator

    def crud(
        self, cmd: CrudCommand
    ) -> Hashable | list[Hashable] | Model | list[Model] | bool | list[bool] | None:
        """
        Generic CRUD operation handler that forwards the command to seqdb while
        setting the functional user.
        """
        casedb_user = cmd.user
        cmd.user = self.seqdb_user
        result = self.seqdb_app.handle(cmd)
        cmd.user = casedb_user
        return result  # type: ignore[no-any-return]



# pylint: disable=wildcard-import, unused-import
# because this is a package, and imported as such in other modules
from __future__ import annotations

import datetime
import uuid
from enum import Enum

import ulid

from gen_epix.commondb.domain.enum import RoleSet as RoleSet


class TimestampFactory(Enum):
    DATETIME_NOW = lambda: datetime.datetime.now()


class IdFactory(Enum):
    UUID4 = uuid.uuid4
    ULID = lambda: ulid.api.new().uuid


class ServiceType(Enum):
    AUTH = "AUTH"
    ORGANIZATION = "ORGANIZATION"
    SYSTEM = "SYSTEM"
    RBAC = "RBAC"
    ABAC = "ABAC"
    OMOP = "OMOP"


class RepositoryType(Enum):
    DICT = "DICT"
    SA_SQLITE = "SA_SQLITE"
    SA_SQL = "SA_SQL"


class Role(Enum):
    ROOT = "OMOPDB_ROOT"
    APP_ADMIN = "OMOPDB_APP_ADMIN"
    ORG_ADMIN = "OMOPDB_ORG_ADMIN"
    REFDATA_ADMIN = "OMOPDB_REFDATA_ADMIN"
    ORG_USER = "OMOPDB_ORG_USER"
    GUEST = "OMOPDB_GUEST"
    ROLE1 = "OMOPDB_ROLE1"


class AnonStrictness(Enum):
    IGNORE = "ignore"
    WARN = "warn"
    STRICT = "strict"


class AnonMethod(Enum):
    MAKE_NULL = "make_null"
    SHIFT = "shift"
    RANDOM = "random"
    CATEGORICAL = "categorical"
    MODEL_ANONYIMIZATION = "model_anonymization"  # for future use



"""
Number to interval transformer implementation.
"""

import math
from collections.abc import Hashable
from decimal import Decimal
from typing import NoReturn, TypedDict

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.enum import NoMatchStrategy
from gen_epix.transform.transformer import Transformer


class IntervalDict(TypedDict):
    """Type definition for interval dictionaries."""

    name: Hashable
    lb: float
    ub: float
    lb_in: bool
    ub_in: bool


class IntervalTransformer(Transformer):
    """
    Maps a number to an interval represented by a hashable, based on the
    bounds of the interval.
    """

    def __init__(
        self,
        src_field: Hashable,
        interval_names: list[Hashable],
        lower_bounds: list[float | int | Decimal | None],
        upper_bounds: list[float | int | Decimal | None],
        tgt_field: Hashable | None = None,
        lower_bound_is_inclusive: list[bool] | bool = True,
        upper_bound_is_inclusive: list[bool] | bool = False,
        name: str | None = None,
        no_match_strategy: NoMatchStrategy = NoMatchStrategy.RAISE,
    ) -> None:

        # Initialise some
        super().__init__(name)
        self.src_field = src_field
        self.tgt_field = tgt_field or src_field
        self._no_match_strategy = no_match_strategy
        self._n_intervals = len(lower_bounds)
        self._lower_bounds = [-math.inf if x is None else x for x in lower_bounds]
        self._upper_bounds = [math.inf if x is None else x for x in upper_bounds]
        self._interval_names = interval_names
        if isinstance(lower_bound_is_inclusive, list):
            self._lower_bound_is_inclusive = list(lower_bound_is_inclusive)
        else:
            self._lower_bound_is_inclusive = [
                lower_bound_is_inclusive
            ] * self._n_intervals
        if isinstance(upper_bound_is_inclusive, list):
            self._upper_bound_is_inclusive = list(upper_bound_is_inclusive)
        else:
            self._upper_bound_is_inclusive = [
                upper_bound_is_inclusive
            ] * self._n_intervals

        # Sort bins
        sorted_indices = sorted(
            range(self._n_intervals), key=lambda i: self._lower_bounds[i]
        )
        self._lower_bounds = [self._lower_bounds[i] for i in sorted_indices]
        self._lower_bound_is_inclusive = [
            self._lower_bound_is_inclusive[i] for i in sorted_indices
        ]
        self._upper_bounds = [self._upper_bounds[i] for i in sorted_indices]
        self._upper_bound_is_inclusive = [
            self._upper_bound_is_inclusive[i] for i in sorted_indices
        ]
        self._interval_names = [self._interval_names[i] for i in sorted_indices]

        # Verify input
        for i in range(self._n_intervals):
            lb = self._lower_bounds[i]
            ub = self._upper_bounds[i]
            if lb > ub:
                raise ValueError(f"Lower bound {lb} must be less than upper bound {ub}")
        for i, lb1 in enumerate(self._lower_bounds[0:-1]):
            lb1_is_inclusive = self._lower_bound_is_inclusive[i]
            ub1 = self._upper_bounds[i]
            ub1_is_inclusive = self._upper_bound_is_inclusive[i]
            lb2 = self._lower_bounds[i + 1]
            ub2 = self._upper_bounds[i + 1]
            lb2_is_inclusive = self._lower_bound_is_inclusive[i + 1]
            ub2_is_inclusive = self._upper_bound_is_inclusive[i + 1]
            if lb2 < ub1 or lb2 == ub1 and (lb2_is_inclusive and ub1_is_inclusive):
                lb1_str = ("[" if lb1_is_inclusive else "]") + str(lb1)
                ub1_str = str(ub1) + ("]" if ub1_is_inclusive else "[")
                lb2_str = ("[" if lb2_is_inclusive else "]") + str(lb2)
                ub2_str = str(ub2) + ("]" if ub2_is_inclusive else "[")
                raise ValueError(
                    f"Intervals overlap: {lb1_str},{ub1_str} and {lb2_str},{ub2_str}"
                )

    def _get_interval(
        self, value: float | int | Decimal | None
    ) -> Hashable | None | NoReturn:
        if value is None:
            return None
        for i in range(self._n_intervals):
            # Match interval
            match_lb = value > self._lower_bounds[i] or (
                value == self._lower_bounds[i] and self._lower_bound_is_inclusive[i]
            )
            match_ub = value < self._upper_bounds[i] or (
                value == self._upper_bounds[i] and self._upper_bound_is_inclusive[i]
            )
            if match_lb and match_ub:
                # Interval matches -> assign value to target field and stop
                return self._interval_names[i]
        return NoReturn

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Map number to interval."""
        src_value = obj.get(self.src_field)
        tgt_value = self._get_interval(src_value)
        if tgt_value == NoReturn:
            if self._no_match_strategy == NoMatchStrategy.RAISE:
                raise ValueError(f"Value {src_value} does not match any interval")
            elif self._no_match_strategy == NoMatchStrategy.SET_NONE:
                obj.set(self.tgt_field, None)
                return obj
            raise NotImplementedError(
                f"Unknown no match strategy {self._no_match_strategy}"
            )
        obj.set(self.tgt_field, tgt_value)
        return obj

    def transform_value(self, value: float | int | Decimal | None) -> Hashable | None:
        """Map number to interval."""
        tgt_value = self._get_interval(value)
        if tgt_value == NoReturn:
            if self._no_match_strategy == NoMatchStrategy.RAISE:
                raise ValueError(f"Value {value} does not match any interval")
            elif self._no_match_strategy == NoMatchStrategy.SET_NONE:
                return None
            raise NotImplementedError(
                f"Unknown no match strategy {self._no_match_strategy}"
            )
        return tgt_value

    def is_transformable(self, value: float | int | Decimal | None) -> bool:
        """Check if a value can be transformed to an interval without performing the transformation."""
        if value is None:
            return True  # None values are always transformable

        tgt_value = self._get_interval(value)
        return tgt_value != NoReturn


class IntervalToIntervalTransformer(Transformer):
    """
    Maps intervals from one categorization to another based on overlapping ranges.

    This transformer takes an interval from a source categorization and maps it to
    the corresponding interval(s) in a target categorization based on range overlaps.
    """

    def __init__(
        self,
        src_field: Hashable,
        src_interval_names: list[Hashable],
        src_lower_bounds: list[float | int | Decimal | None],
        src_upper_bounds: list[float | int | Decimal | None],
        tgt_interval_names: list[Hashable],
        tgt_lower_bounds: list[float | int | Decimal | None],
        tgt_upper_bounds: list[float | int | Decimal | None],
        tgt_field: Hashable | None = None,
        src_lower_bound_is_inclusive: list[bool] | bool = True,
        src_upper_bound_is_inclusive: list[bool] | bool = False,
        tgt_lower_bound_is_inclusive: list[bool] | bool = True,
        tgt_upper_bound_is_inclusive: list[bool] | bool = False,
        name: str | None = None,
        no_match_strategy: NoMatchStrategy = NoMatchStrategy.RAISE,
        overlap_strategy: str = "largest_overlap",  # "largest_overlap" or "exact_fit"
    ) -> None:

        super().__init__(name)
        self.src_field = src_field
        self.tgt_field = tgt_field or src_field
        self._no_match_strategy = no_match_strategy
        self._overlap_strategy = overlap_strategy

        # Initialize source intervals
        self._src_intervals = self._create_interval_list(
            src_interval_names,
            src_lower_bounds,
            src_upper_bounds,
            src_lower_bound_is_inclusive,
            src_upper_bound_is_inclusive,
        )

        # Initialize target intervals
        self._tgt_intervals = self._create_interval_list(
            tgt_interval_names,
            tgt_lower_bounds,
            tgt_upper_bounds,
            tgt_lower_bound_is_inclusive,
            tgt_upper_bound_is_inclusive,
        )

        # Pre-compute mapping from source to target intervals
        self._interval_mapping = self._compute_interval_mapping()

    def _create_interval_list(
        self,
        names: list[Hashable],
        lower_bounds: list[float | int | Decimal | None],
        upper_bounds: list[float | int | Decimal | None],
        lower_inclusive: list[bool] | bool,
        upper_inclusive: list[bool] | bool,
    ) -> list[IntervalDict]:
        """Create standardized interval representation."""
        n_intervals = len(names)

        if isinstance(lower_inclusive, bool):
            lower_inclusive = [lower_inclusive] * n_intervals
        if isinstance(upper_inclusive, bool):
            upper_inclusive = [upper_inclusive] * n_intervals

        intervals = []
        for i in range(n_intervals):
            intervals.append(
                IntervalDict(
                    name=names[i],
                    lb=(
                        -math.inf
                        if lower_bounds[i] is None
                        else float(lower_bounds[i] or 0)
                    ),
                    ub=(
                        math.inf
                        if upper_bounds[i] is None
                        else float(upper_bounds[i] or 0)
                    ),
                    lb_in=lower_inclusive[i],
                    ub_in=upper_inclusive[i],
                )
            )

        return intervals

    def _compute_interval_mapping(self) -> dict[Hashable, Hashable]:
        """Pre-compute mapping from source intervals to target intervals."""
        mapping = {}

        for src_interval in self._src_intervals:
            best_match = None
            max_overlap = 0

            for tgt_interval in self._tgt_intervals:
                overlap = self._calculate_overlap(src_interval, tgt_interval)

                if self._overlap_strategy == "exact_fit":
                    # Only map if source interval is completely contained in target
                    if self._is_contained(src_interval, tgt_interval):
                        mapping[src_interval["name"]] = tgt_interval["name"]
                        break
                elif self._overlap_strategy == "largest_overlap":
                    # Map to target with largest overlap
                    if overlap > max_overlap:
                        max_overlap = overlap
                        best_match = tgt_interval["name"]

            if (
                self._overlap_strategy == "largest_overlap"
                and best_match
                and max_overlap > 0
            ):
                mapping[src_interval["name"]] = best_match

        return mapping

    def _calculate_overlap(
        self,
        src_interval: IntervalDict,
        tgt_interval: IntervalDict,
    ) -> float:
        """Calculate overlap between two intervals."""
        # Determine effective bounds for overlap calculation
        overlap_start = max(src_interval["lb"], tgt_interval["lb"])
        overlap_end = min(src_interval["ub"], tgt_interval["ub"])

        if overlap_start >= overlap_end:
            return 0.0

        # Handle infinite bounds
        if math.isinf(overlap_start) or math.isinf(overlap_end):
            return float("inf")

        return overlap_end - overlap_start

    def _is_contained(
        self, src_interval: IntervalDict, tgt_interval: IntervalDict
    ) -> bool:
        """Check if source interval is completely contained in target interval."""
        # Check lower bound
        if src_interval["lb"] < tgt_interval["lb"]:
            return False
        if (
            src_interval["lb"] == tgt_interval["lb"]
            and src_interval["lb_in"]
            and not tgt_interval["lb_in"]
        ):
            return False

        # Check upper bound
        if src_interval["ub"] > tgt_interval["ub"]:
            return False
        if (
            src_interval["ub"] == tgt_interval["ub"]
            and src_interval["ub_in"]
            and not tgt_interval["ub_in"]
        ):
            return False

        return True

    def _map_interval(self, src_interval_name: Hashable) -> Hashable | None | NoReturn:
        """Map source interval name to target interval name."""
        if src_interval_name is None:
            return None

        mapped_name = self._interval_mapping.get(src_interval_name)
        if mapped_name is None:
            return NoReturn

        return mapped_name

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Transform interval from source categorization to target categorization."""
        src_value = obj.get(self.src_field)
        tgt_value = self._map_interval(src_value)

        if tgt_value == NoReturn:
            if self._no_match_strategy == NoMatchStrategy.RAISE:
                raise ValueError(
                    f"Interval {src_value} cannot be mapped to target categorization"
                )
            elif self._no_match_strategy == NoMatchStrategy.SET_NONE:
                obj.set(self.tgt_field, None)
                return obj
            raise NotImplementedError(
                f"Unknown no match strategy {self._no_match_strategy}"
            )

        obj.set(self.tgt_field, tgt_value)
        return obj

    def transform_value(self, src_interval_name: Hashable) -> Hashable | None:
        """Transform interval name directly."""
        tgt_value = self._map_interval(src_interval_name)

        if tgt_value == NoReturn:
            if self._no_match_strategy == NoMatchStrategy.RAISE:
                raise ValueError(
                    f"Interval {src_interval_name} cannot be mapped to target categorization"
                )
            elif self._no_match_strategy == NoMatchStrategy.SET_NONE:
                return None
            raise NotImplementedError(
                f"Unknown no match strategy {self._no_match_strategy}"
            )

        return tgt_value

    def is_transformable(self, src_interval_name: Hashable) -> bool:
        """Check if source interval can be mapped to target categorization."""
        if src_interval_name is None:
            return True

        return src_interval_name in self._interval_mapping



from pydantic import BaseModel


class Link(BaseModel, frozen=True):
    """
    Represents a link between entities.

    Attributes
    ----------
    link_field_name : str
        The name of the field that represents the link.
    link_model_class : type[BaseModel]
        The model class that the link points to.
    relationship_field_name : str, optional
        The name of the field used for back-population, by default None.

    Methods
    -------
    None
    """

    link_field_name: str
    link_model_class: type[BaseModel]
    relationship_field_name: str | None = None

    def to_tuple(self) -> tuple[str, type[BaseModel], str | None]:
        return (
            self.link_field_name,
            self.link_model_class,
            self.relationship_field_name,
        )

    @classmethod
    def from_tuple(cls, tuple_: tuple[str, type[BaseModel], str | None]) -> "Link":
        return cls(
            link_field_name=tuple_[0],
            link_model_class=tuple_[1],
            relationship_field_name=tuple_[2],
        )



from typing import ClassVar, Self
from uuid import UUID

from pydantic import Field, model_validator

from gen_epix.commondb.domain.model import Model
from gen_epix.fastapp import Entity
from gen_epix.seqdb.domain import enum
from gen_epix.seqdb.domain.model.seq.metadata import RefSnp
from gen_epix.seqdb.domain.model.seq.persistable import (
    AstMeasurement,
    AstPrediction,
    PcrMeasurement,
    ReadSet,
    SeqClassification,
    SeqDistanceProtocol,
    SeqTaxonomy,
)


class CompleteContig(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_contigs",
        persistable=False,
    )
    seq_id: UUID = Field(description="The ID of the sequence.")
    seq: str = Field(description="The contig sequence.")
    qc: enum.QualityControlResult = Field(
        description="The quality control result of the contig sequence."
    )
    index: int = Field(description="The index of the contig in the sequence.")


class CompleteAlleleProfile(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_allele_profiles",
        persistable=False,
    )
    seq_id: UUID = Field(description="The ID of the sequence.")
    locus_set_id: UUID = Field(description="The ID of the locus set.")
    locus_ids: list[UUID] = Field(description="The IDs of the loci.")
    allele_ids: list[UUID | None] = Field(
        description="The IDs of the alleles for each locus."
    )
    multiple_allele_ids: dict[UUID, list[UUID]] = Field(
        description="Mapping of locus ID to multiple allele IDs."
    )
    allele_count_by_qc: dict[enum.QualityControlResult, int] = Field(
        description="Mapping of quality control result to allele count."
    )


class CompleteSnpProfile(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_snp_profiles",
        persistable=False,
    )
    seq_id: UUID = Field(description="The ID of the sequence.")
    ref_snps: list[RefSnp] = Field(description="The list of reference SNPs.")
    snps: str = Field(description="The SNPs in the profile.")
    snp_profile: str = Field(description="The SNP profile string.")
    snp_profile_format: enum.SnpProfileFormat = Field(
        description="The format of the SNP profile."
    )


class CompleteSeq(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_seqs",
        persistable=False,
    )
    sample_id: UUID | None = Field(default=None, description="The ID of the sample.")
    primary_taxon_id: UUID | None = Field(
        default=None, description="The ID of the primary taxon."
    )
    read_sets: list[ReadSet] | None = Field(
        default=None, description="The list of read sets associated with the sequence."
    )
    contigs: list[CompleteContig] | None = Field(
        default=None, description="The list of contigs associated with the sequence."
    )
    allele_profiles: list[CompleteAlleleProfile] | None = Field(
        default=None,
        description="The list of allele profiles associated with the sequence.",
    )
    snp_profiles: list[CompleteSnpProfile] | None = Field(
        default=None,
        description="The list of SNP profiles associated with the sequence.",
    )
    taxa: list[SeqTaxonomy] | None = Field(
        default=None, description="The list of taxonomies associated with the sequence."
    )
    ast_predictions: list[AstPrediction] | None = Field(
        default=None,
        description="The list of AST predictions associated with the sequence.",
    )
    classifications: list[SeqClassification] | None = Field(
        default=None,
        description="The list of classifications associated with the sequence.",
    )
    qc: enum.QualityControlResult | None = Field(
        default=None, description="The quality control result of the sequence."
    )


class CompleteSample(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_samples",
        persistable=False,
    )
    primary_seq_id: UUID | None = Field(
        default=None, description="The ID of the primary sequence."
    )
    primary_taxon_id: UUID | None = Field(
        default=None, description="The ID of the primary taxon."
    )
    seqs: list[CompleteSeq] | None = Field(
        default=None, description="The list of sequences associated with the sample."
    )
    pcr_measurements: list[PcrMeasurement] | None = Field(
        default=None,
        description="The list of PCR measurements associated with the sample.",
    )
    ast_measurements: list[AstMeasurement] | None = Field(
        default=None,
        description="The list of AST measurements associated with the sample.",
    )


class PhylogeneticTree(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="phylogenetic_trees",
        persistable=False,
    )
    tree_algorithm: enum.TreeAlgorithm = Field(description="The tree algorithm")
    seq_distance_protocol_id: UUID = Field(
        description="The ID of the sequence distance protocol. FOREIGN KEY"
    )
    seq_distance_protocol: SeqDistanceProtocol | None = Field(
        default=None, description="The sequence distance protocol"
    )
    leaf_names: list[str] | None = Field(
        default=None,
        description="The list of names of the leaves of the phylogenetic tree to be put in the tree representation instead of seq_ids. Must have the same length as seq_ids.",
    )
    seq_ids: list[UUID] | None = Field(
        default=None,
        description="The list of unique identifiers of the sequence of each leaf of the phylogenetic tree.",
    )
    newick_repr: str = Field(
        description="The Newick representation of the phylogenetic tree."
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if self.leaf_names:
            if len(set(self.leaf_names)) < len(self.leaf_names):
                raise ValueError("Duplicate leaf_codes")
        if self.seq_ids:
            if len(set(self.seq_ids)) < len(self.seq_ids):
                raise ValueError("Duplicate seq_ids")
            if self.leaf_names and len(self.seq_ids) != len(self.leaf_names):
                raise ValueError(
                    "seq_ids and leaf_codes must have the same length if leaf_codes is provided."
                )
        return self


class MultipleAlignment(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="multiple_alignments",
        persistable=False,
    )
    alignment_protocol_id: UUID = Field(
        description="The ID of the alignment protocol. FOREIGN KEY"
    )
    seq_ids: list[UUID] = Field(
        description="The list of sequence IDs included in the multiple alignment."
    )
    n_seqs: int = Field(description="The number of sequences in the alignment.")
    n_contigs: list[int] = Field(
        description="The number of contigs for each sequence in the alignment."
    )
    contig_seqs: list[list[str]] = Field(
        description="The list of contig sequences for each sequence in the alignment."
    )
    n_alignments: int = Field(description="The number of alignments.")
    n_columns: list[int] = Field(
        description="The number of columns for each alignment."
    )
    start_columns: list[list[int]] = Field(
        description="The start column positions for each alignment."
    )
    contig_ordinals: list[list[int]] = Field(
        description="The ordinals of the contigs for each sequence in the alignment."
    )
    contig_start_positions: list[list[int]] = Field(
        description="The start positions of the contigs for each sequence in the alignment."
    )
    contig_directions: list[list[bool]] = Field(
        description="The directions of the contigs for each sequence in the alignment."
    )
    lengths: list[list[int]] = Field(
        description="The lengths of the contigs for each sequence in the alignment."
    )



from typing import Any

from gen_epix.commondb.domain import command
from gen_epix.commondb.domain.service.system import BaseSystemService
from gen_epix.fastapp import PermissionType, Policy


class BaseHasSystemOutagePolicy(Policy):
    """
    Policy that checks if the system has a current outage

    """

    def __init__(
        self,
        system_service: BaseSystemService,
        **kwargs: Any,
    ):
        self.system_service = system_service
        self.props = kwargs
        self.outage_update_permission = system_service.app.domain.get_permission(
            command.OutageCrudCommand, PermissionType.UPDATE
        )



# pylint: disable=useless-import-alias
from gen_epix import fastapp
from gen_epix.commondb.domain import command as common_command
from gen_epix.commondb.domain import enum as common_enum
from gen_epix.commondb.domain.command import (
    COMMANDS_BY_SERVICE_TYPE as _COMMON_COMMANDS_BY_SERVICE_TYPE,
)
from gen_epix.commondb.domain.command import Command as Command
from gen_epix.commondb.domain.command import ContactCrudCommand as ContactCrudCommand
from gen_epix.commondb.domain.command import CrudCommand as CrudCommand
from gen_epix.commondb.domain.command import (
    DataCollectionCrudCommand as DataCollectionCrudCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetCrudCommand as DataCollectionSetCrudCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetDataCollectionUpdateAssociationCommand as DataCollectionSetDataCollectionUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetMemberCrudCommand as DataCollectionSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command import (
    GetIdentityProvidersCommand as GetIdentityProvidersCommand,
)
from gen_epix.commondb.domain.command import (
    IdentifierIssuerCrudCommand as IdentifierIssuerCrudCommand,
)
from gen_epix.commondb.domain.command import InviteUserCommand as InviteUserCommand
from gen_epix.commondb.domain.command import (
    OrganizationAdminPolicyCrudCommand as OrganizationAdminPolicyCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationCrudCommand as OrganizationCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetCrudCommand as OrganizationSetCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetMemberCrudCommand as OrganizationSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetOrganizationUpdateAssociationCommand as OrganizationSetOrganizationUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import OutageCrudCommand as OutageCrudCommand
from gen_epix.commondb.domain.command import (
    RegisterInvitedUserCommand as RegisterInvitedUserCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOrganizationAdminNameEmailsCommand as RetrieveOrganizationAdminNameEmailsCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOrganizationContactCommand as RetrieveOrganizationContactCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOutagesCommand as RetrieveOutagesCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOwnPermissionsCommand as RetrieveOwnPermissionsCommand,
)
from gen_epix.commondb.domain.command import SiteCrudCommand as SiteCrudCommand
from gen_epix.commondb.domain.command import (
    UpdateAssociationCommand as UpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import UpdateUserCommand as UpdateUserCommand
from gen_epix.commondb.domain.command import (
    UpdateUserOwnOrganizationCommand as UpdateUserOwnOrganizationCommand,
)
from gen_epix.commondb.domain.command import UserCrudCommand as UserCrudCommand
from gen_epix.commondb.domain.command import (
    UserInvitationCrudCommand as UserInvitationCrudCommand,
)
from gen_epix.commondb.domain.command.abac import (
    RetrieveOrganizationsUnderAdminCommand as RetrieveOrganizationsUnderAdminCommand,
)
from gen_epix.commondb.domain.command.organization import (
    RetrieveInviteUserConstraintsCommand as RetrieveInviteUserConstraintsCommand,
)
from gen_epix.commondb.domain.command.rbac import (
    RetrieveSubRolesCommand as RetrieveSubRolesCommand,
)
from gen_epix.seqdb.domain import enum
from gen_epix.seqdb.domain.command.file import FileCrudCommand as FileCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    AlignmentProtocolCrudCommand as AlignmentProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    AlleleAlignmentCrudCommand as AlleleAlignmentCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import AlleleCrudCommand as AlleleCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    AlleleProfileCrudCommand as AlleleProfileCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    AssemblyProtocolCrudCommand as AssemblyProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    AstMeasurementCrudCommand as AstMeasurementCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    AstPredictionCrudCommand as AstPredictionCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    AstProtocolCrudCommand as AstProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    GenerateMultipleAlignmentCommand as GenerateMultipleAlignmentCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    GeneratePhylogeneticTreeCommand as GeneratePhylogeneticTreeCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    KmerDetectionProtocolCrudCommand as KmerDetectionProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    KmerProfileCrudCommand as KmerProfileCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    LibraryPrepProtocolCrudCommand as LibraryPrepProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import LocusCrudCommand as LocusCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    LocusDetectionProtocolCrudCommand as LocusDetectionProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import LocusSetCrudCommand as LocusSetCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    LocusSetMemberCrudCommand as LocusSetMemberCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    PcrMeasurementCrudCommand as PcrMeasurementCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    PcrProtocolCrudCommand as PcrProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import RawSeqCrudCommand as RawSeqCrudCommand
from gen_epix.seqdb.domain.command.seq import ReadSetCrudCommand as ReadSetCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    RefAlleleCrudCommand as RefAlleleCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import RefSeqCrudCommand as RefSeqCrudCommand
from gen_epix.seqdb.domain.command.seq import RefSnpCrudCommand as RefSnpCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    RefSnpSetCrudCommand as RefSnpSetCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RefSnpSetMemberCrudCommand as RefSnpSetMemberCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrieveCompleteAlleleProfileCommand as RetrieveCompleteAlleleProfileCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrieveCompleteContigCommand as RetrieveCompleteContigCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrieveCompleteSampleCommand as RetrieveCompleteSampleCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrieveCompleteSeqCommand as RetrieveCompleteSeqCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrieveCompleteSnpProfileCommand as RetrieveCompleteSnpProfileCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrieveMultipleAlignmentCommand as RetrieveMultipleAlignmentCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrievePhylogeneticTreeCommand as RetrievePhylogeneticTreeCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    RetrieveSeqFastaCommand as RetrieveSeqFastaCommand,
)
from gen_epix.seqdb.domain.command.seq import SampleCrudCommand as SampleCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    SeqAlignmentCrudCommand as SeqAlignmentCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SeqCategoryCrudCommand as SeqCategoryCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SeqCategorySetCrudCommand as SeqCategorySetCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SeqClassificationCrudCommand as SeqClassificationCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SeqClassificationProtocolCrudCommand as SeqClassificationProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import SeqCrudCommand as SeqCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    SeqDistanceCrudCommand as SeqDistanceCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SeqDistanceProtocolCrudCommand as SeqDistanceProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SeqTaxonomyCrudCommand as SeqTaxonomyCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SnpDetectionProtocolCrudCommand as SnpDetectionProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SnpProfileCrudCommand as SnpProfileCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    SubtypingSchemeCrudCommand as SubtypingSchemeCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import TaxonCrudCommand as TaxonCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    TaxonLocusLinkCrudCommand as TaxonLocusLinkCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    TaxonomyProtocolCrudCommand as TaxonomyProtocolCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import TaxonSetCrudCommand as TaxonSetCrudCommand
from gen_epix.seqdb.domain.command.seq import (
    TaxonSetMemberCrudCommand as TaxonSetMemberCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    TreeAlgorithmClassCrudCommand as TreeAlgorithmClassCrudCommand,
)
from gen_epix.seqdb.domain.command.seq import (
    TreeAlgorithmCrudCommand as TreeAlgorithmCrudCommand,
)

COMMANDS_BY_SERVICE_TYPE: dict[enum.ServiceType, set[type[fastapp.Command]]] = {
    # Specific commands
    enum.ServiceType.ABAC: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.ABAC]
    )
    | set(),
    enum.ServiceType.SEQ: {
        AlignmentProtocolCrudCommand,
        AlleleAlignmentCrudCommand,
        AlleleCrudCommand,
        AlleleProfileCrudCommand,
        AssemblyProtocolCrudCommand,
        AstMeasurementCrudCommand,
        AstPredictionCrudCommand,
        AstProtocolCrudCommand,
        GenerateMultipleAlignmentCommand,
        GeneratePhylogeneticTreeCommand,
        KmerDetectionProtocolCrudCommand,
        KmerProfileCrudCommand,
        LibraryPrepProtocolCrudCommand,
        LocusCrudCommand,
        LocusDetectionProtocolCrudCommand,
        LocusSetCrudCommand,
        LocusSetMemberCrudCommand,
        PcrMeasurementCrudCommand,
        PcrProtocolCrudCommand,
        RawSeqCrudCommand,
        ReadSetCrudCommand,
        RefAlleleCrudCommand,
        RefSeqCrudCommand,
        RefSnpCrudCommand,
        RefSnpSetCrudCommand,
        RefSnpSetMemberCrudCommand,
        RetrieveCompleteAlleleProfileCommand,
        RetrieveCompleteContigCommand,
        RetrieveCompleteSampleCommand,
        RetrieveCompleteSeqCommand,
        RetrieveCompleteSnpProfileCommand,
        RetrieveMultipleAlignmentCommand,
        RetrievePhylogeneticTreeCommand,
        RetrieveSeqFastaCommand,
        SampleCrudCommand,
        SeqAlignmentCrudCommand,
        SeqCategoryCrudCommand,
        SeqCategorySetCrudCommand,
        SeqClassificationCrudCommand,
        SeqClassificationProtocolCrudCommand,
        SeqCrudCommand,
        SeqDistanceCrudCommand,
        SeqDistanceProtocolCrudCommand,
        SeqTaxonomyCrudCommand,
        SnpDetectionProtocolCrudCommand,
        SnpProfileCrudCommand,
        SubtypingSchemeCrudCommand,
        TaxonCrudCommand,
        TaxonLocusLinkCrudCommand,
        TaxonomyProtocolCrudCommand,
        TaxonSetCrudCommand,
        TaxonSetMemberCrudCommand,
        TreeAlgorithmClassCrudCommand,
        TreeAlgorithmCrudCommand,
    },
    enum.ServiceType.FILE: {
        FileCrudCommand,
    },
    # Common commands
    enum.ServiceType.AUTH: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.AUTH]
    ),
    enum.ServiceType.SYSTEM: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.SYSTEM]
    ),
    enum.ServiceType.RBAC: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.RBAC]
    ),
    enum.ServiceType.ORGANIZATION: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.ORGANIZATION]
    ),
}

COMMON_COMMAND_MAP: dict[type[fastapp.Command], type[fastapp.Command]] = {
    common_command.UserCrudCommand: UserCrudCommand,
    common_command.UserInvitationCrudCommand: UserInvitationCrudCommand,
    common_command.InviteUserCommand: InviteUserCommand,
    common_command.UpdateUserCommand: UpdateUserCommand,
    common_command.OrganizationAdminPolicyCrudCommand: OrganizationAdminPolicyCrudCommand,
}



from typing import Any

from gen_epix.commondb.domain.service.rbac import BaseRbacService
from gen_epix.fastapp import Policy


class BaseIsPermissionSubsetNewRolePolicy(Policy):
    """
    Policy that checks if the user has the required permissions to create or update a
    role.

    The user must have all the permissions that the new role has to avoid elevation of
    privileges.

    Does not apply to read or delete operations.
    """

    def __init__(self, rbac_service: BaseRbacService, **kwargs: Any):
        self.rbac_service = rbac_service
        self.props = kwargs



from datetime import datetime
from typing import ClassVar
from uuid import UUID

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, field_serializer

from gen_epix import fastapp
from gen_epix.casedb.domain import enum
from gen_epix.casedb.domain.model.case.persistable import Case
from gen_epix.casedb.domain.model.seqdb import ReadSet as ReadSet
from gen_epix.casedb.domain.model.seqdb import Seq as Seq
from gen_epix.commondb.domain.model import Model
from gen_epix.commondb.util import copy_model_field
from gen_epix.fastapp.domain import Entity
from gen_epix.filter import TypedCompositeFilter, TypedDatetimeRangeFilter


class CaseForCreateUpdate(Model):
    """
    A class representing a case to be created or updated.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="cases_for_create_update",
        persistable=False,
    )
    subject_id: UUID | None = copy_model_field(Case, "subject_id")
    count: int | None = copy_model_field(Case, "count")
    case_date: datetime | None = Field(
        description="The date of the case. Required when creating a case, ignored when updating.",
        default=None,
    )
    content: dict[UUID, str | None] = Field(
        description="The column data of the case as {col_id: str_value}. If None and the model is used for update, then any existing value will be deleted."
    )

    @field_serializer("content", mode="plain")
    def _serialize_content(
        self, value: dict[UUID, str | None]
    ) -> dict[str, str | None]:
        return {str(x): y for x, y in value.items()}


class CaseTypeDim(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_dims",
        persistable=False,
    )
    id: UUID = Field(description="The ID of the first case type column.")
    dim_id: UUID = Field(description="The ID of the dimension. FOREIGN KEY")
    occurrence: int | None = Field(
        default=None,
        description=(
            "The index of the occurrence of the dimension for this case type."
            " E.g. for first and second vaccination time it would be 1 and 2."
            " Empty if only a single occurrence."
        ),
    )
    rank: int = Field(description="The rank of the case type dimension for ordering.")
    case_type_col_order: list[UUID] = Field(
        description="The order of the case type columns"
    )


class CaseTypeStat(fastapp.Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_type_stats",
        persistable=False,
    )
    case_type_id: UUID = Field(description="The ID of the case type.")
    n_cases: int | None = Field(
        default=None, description="The number of cases for the case type."
    )
    first_case_month: str | None = Field(
        default=None, description="The ISO year and month of the first case."
    )
    last_case_month: str | None = Field(
        default=None, description="The ISO year and month of the last case."
    )


class CaseSetStat(fastapp.Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_set_stats",
        persistable=False,
    )
    case_set_id: UUID = Field(description="The ID of the case set.")
    n_cases: int | None = Field(
        default=None, description="The number of cases in the case set."
    )
    n_own_cases: int | None = Field(
        default=None, description="The number of own cases in the case set."
    )
    first_case_month: str | None = Field(
        default=None, description="The ISO year and month of the first case."
    )
    last_case_month: str | None = Field(
        default=None, description="The ISO year and month of the last case."
    )


class CaseQuery(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_queries",
        persistable=False,
    )
    label: str | None = Field(default=None, description="The label for the query.")
    case_type_ids: set[UUID] | None = Field(
        default=None,
        description="The IDs of the case type(s) that the case must belong to. Not applied if not provided.",
    )
    case_set_ids: set[UUID] | None = Field(
        default=None,
        description="The IDs of the case set(s) that the case must belong to. Not applied if not provided.",
    )
    datetime_range_filter: TypedDatetimeRangeFilter | None = Field(
        default=None,
        description="The datetime range filter to apply to the case date. Not applied if not provided.",
    )
    # TODO: add data_collection_id
    filter: TypedCompositeFilter | None = Field(
        default=None, description="The filter to apply. Not applied if not provided."
    )


class CaseSetQuery(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_set_queries",
        persistable=False,
    )
    label: str = Field(description="The label for the query.")
    filter: TypedCompositeFilter = Field(description="The filter to apply.")


class BaseCaseRights(Model):
    created_in_data_collection_id: UUID = Field(
        description="The ID of the data collection where the item was created",
    )
    case_type_id: UUID = Field(description="The ID of the case type")
    data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections in which the item is currently shared, including the created_in_data_collection_id",
    )
    is_full_access: bool = Field(
        description="Whether the user has full access to the item, i.e. all rights on all data collections",
    )
    add_data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections to which the item is allowed to be added",
    )
    remove_data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections from which the item is allowed to be removed. If remove_data_collection_ids is equal to data_collection_ids, the item is allowed to be deleted",
    )
    can_delete: bool = Field(
        description="Whether the item can be deleted.",
    )
    shared_in_data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections in which the item is currently shared, excluding the created_in_data_collection_id",
    )


class CaseRights(BaseCaseRights):
    """
    Describes all the rights that a user has on one particular case, based on the data
    collections in which it is currently shared.
    """

    NAME: ClassVar = "CaseRights"
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_rights",
        persistable=False,
    )
    case_id: UUID = Field(description="The ID of the case")
    read_case_type_col_ids: set[UUID] = Field(
        description="The IDs of the case type columns that are allowed to be read for the case",
    )
    write_case_type_col_ids: set[UUID] = Field(
        description="The IDs of the case type columns that are allowed to be written for the case",
    )


class CaseSetRights(BaseCaseRights):
    """
    Describes all the rights that a user has on one particular case set, based on the
    data collections in which it is currently shared.
    """

    NAME: ClassVar = "CaseSetRights"
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_set_rights",
        persistable=False,
    )
    case_set_id: UUID = Field(description="The ID of the case set")
    read_case_set: bool = Field(
        description="Whether the case set is allowed to be read",
    )
    write_case_set: bool = Field(
        description="Whether the case set is allowed to be written",
    )


class CaseDataIssue(PydanticBaseModel):
    case_type_col_id: UUID = Field(description="The ID of the case type column")
    original_value: str | None = Field(description="The value of the case type column")
    updated_value: str | None = Field(
        description="The new value of the case type column after potential resolution. If not resolved, this will be None.",
    )
    data_rule: enum.CaseColDataRule = Field(description="The type of validation issue")
    details: str | None = Field(description="The details of the data issue")


class ValidatedCase(PydanticBaseModel):
    case: CaseForCreateUpdate = Field(description="The case with validated content.")
    data_issues: list[CaseDataIssue] = Field(
        description="The data issues found for the case."
    )


class CaseValidationReport(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_validation_reports",
        persistable=False,
    )
    case_type_id: UUID = Field(description="The case type ID that the cases belong to.")
    created_in_data_collection_id: UUID = Field(
        description="The data collection ID in which the cases would be created."
    )
    is_update: bool = Field(
        description="Whether the cases are intended to be updated or newly created."
    )
    data_collection_ids: set[UUID] = Field(
        description="The additional data collections that the cases would be put in, other than the created_in_data_collection."
    )
    validated_cases: list[ValidatedCase] = Field(
        description="The cases containing validated content and any data issues found during validation."
    )


class CaseReadSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_read_sets",
        persistable=False,
    )
    case_id: UUID = Field(
        description="The ID of the case that the read set is or will be associated with."
    )
    case_type_col_id: UUID = Field(
        description="The ID of the case type column that the read set is or will be associated with."
    )
    read_set_id: UUID | None = Field(
        description="The ID of the read set.", default=None
    )
    read_set: ReadSet | None = Field(default=None, description="The read set.")


class CaseSeq(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="case_seq",
        persistable=False,
    )
    case_id: UUID = Field(
        description="The ID of the case that the sequence is or will be associated with."
    )
    case_type_col_id: UUID = Field(
        description="The ID of the case type column that the sequence is or will be associated with."
    )
    seq_id: UUID | None = Field(description="The ID of the sequence.", default=None)
    seq: Seq | None = Field(default=None, description="The sequence.")



import datetime
import gzip
import importlib
import logging
import os
import pickle
import tomllib
import uuid
from collections.abc import Hashable, Iterable
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Any

import ulid
from pydantic import BaseModel, Field

from gen_epix.commondb.config import AppCfg
from gen_epix.commondb.domain.enum import (
    AppType,
    AppTypeSet,
    DevIdpConfig,
    DevRepositoryConfig,
    DevRepositoryConfigSet,
)
from gen_epix.fastapp import Command, Domain, Model, exc
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.fastapp.repositories.dict import DictRepository
from gen_epix.fastapp.repositories.sa import SARepository


def generate_ulid() -> uuid.UUID:
    return ulid.api.new().uuid


def get_project_root() -> Path:
    """
    Get the root path of the project by looking for pyproject.toml.

    Searches upward from the current file's directory until it finds
    a directory containing pyproject.toml, which indicates the project root.

    Returns:
        Path: The absolute path to the project root directory.

    Raises:
        FileNotFoundError: If pyproject.toml cannot be found in any parent directory.
    """
    current_dir = Path(__file__).parent

    while current_dir != current_dir.parent:
        if (current_dir / "pyproject.toml").exists():
            return current_dir.resolve()
        current_dir = current_dir.parent

    raise FileNotFoundError("Could not find pyproject.toml in any parent directory")


def map_paired_elements(
    data: Iterable[tuple[Hashable, Any]], as_set: bool = False, frozen: bool = False
) -> (
    dict[Hashable, list[Any]]
    | dict[Hashable, set[Any]]
    | dict[Hashable, frozenset[Any]]
):
    """
    Convert an iterable of paired elements to a dictionary of lists or sets, where
    the keys are the unique first elements and the values the list or set of second
    elements matching that key in the input. If frozen=True, the sets are converted
    to frozensets.
    """
    retval: (
        dict[Hashable, list[Any]]
        | dict[Hashable, set[Any]]
        | dict[Hashable, frozenset[Any]]
    ) = {}
    if as_set:
        for k, v in data:
            if k not in retval:
                retval[k] = set()  # type: ignore[assignment]
            retval[k].add(v)  # type: ignore[union-attr]
        if frozen:
            for k in retval:
                retval[k] = frozenset(retval[k])  # type: ignore[assignment]
    else:
        for k, v in data:
            if k not in retval:
                retval[k] = []  # type: ignore[assignment]
            retval[k].append(v)  # type: ignore[union-attr]
    return retval


# Get version with fallback for development
@lru_cache(maxsize=1)
def get_package_version() -> str:
    """Retrieve the project version from the pyproject.toml file.
    Must be run from the project root directory.

    Returns:
        str: The version of the project as specified in pyproject.toml.
    """
    pyproject_path = "pyproject.toml"

    with open(pyproject_path, "rb") as f:
        pyproject_data = tomllib.load(f)

    return pyproject_data["project"]["version"]


def register_domain_entities(
    domain: Domain,
    sorted_service_types: Iterable[Hashable],
    sorted_models_by_service_type: dict[Hashable, list[type[Model]]],
    commands_by_service_type: dict[Hashable, set[type[Command]]],
    common_model_map: dict[type[Model], type[Model]] | None = None,
    common_command_map: dict[type[Command], type[Command]] | None = None,
    set_schema_to_service_type: bool = False,
) -> None:
    """
    Register service types, models and commands with a domain. In case some
    models or commands are subclassed from another domain and the provides
    models and commands contain their parent classes, they can be substituted
    in the input and subsequently be registered as the actual classes, by
    providing a mapping.

    If `set_schema_to_service_type` is enabled, the schema name of the model
    will be set to the lower case service name for persistable entities, unless
    the schema name is already set.
    """
    if not common_model_map:
        common_model_map = {}
    for service_type in sorted_service_types:
        # Register the service type
        domain.register_service_type(service_type)
        schema_name = (
            str(service_type.value).lower()
            if isinstance(service_type, Enum)
            else str(service_type)
        )
        # Register the models
        for i, model_class in enumerate(
            sorted_models_by_service_type.get(service_type, [])
        ):
            if model_class in common_model_map:
                # Substitute the model class with its commondb implementation,
                # also in the input
                model_class = common_model_map[model_class]
                sorted_models_by_service_type[service_type][i] = model_class
            if model_class.ENTITY is None:
                raise exc.InitializationServiceError(
                    f"Entity for model class {model_class} is not initialized."
                )
            if (
                set_schema_to_service_type
                and model_class.ENTITY.persistable
                and model_class.ENTITY.schema_name is None
            ):
                model_class.ENTITY.schema_name = schema_name
            domain.register_entity(
                model_class.ENTITY, model_class=model_class, service_type=service_type
            )
        # Register the commands
        for command_class in commands_by_service_type.get(service_type, []):
            if common_command_map and command_class in common_command_map:
                # Substitute the command class with its commondb implementation,
                # also in the input
                commands_by_service_type[service_type].remove(command_class)
                command_class = common_command_map[command_class]
                commands_by_service_type[service_type].add(command_class)
            domain.register_command(command_class, service_type=service_type)


def copy_model_field(
    from_model: type[BaseModel], field_name: str, **kwargs: Any
) -> Any:
    """
    Copy a field from a model
    """
    field_info_attributes = {
        "alias": "alias",
        "alias_priority": "alias_priority",
        "default": "default",
        "default_factory": "default_factory",
        "deprecated": "deprecated",
        "description": "description",
        "discriminator": "discriminator",
        "examples": "examples",
        "exclude": "exclude",
        "frozen": "frozen",
        "init": "init",
        "init_var": "init_var",
        "json_schema_extra": "json_schema_extra",
        "kw_only": "kw_only",
        "serialization_alias": "serialization_alias",
        "title": "title",
    }
    metadata_attributes = {
        "allow_inf_nan": "allow_inf_nan",
        "coerce_numbers_to_str": "coerce_numbers_to_str",
        "decimal_places": "decimal_places",
        "ge": "ge",
        "gt": "gt",
        "le": "le",
        "lt": "lt",
        "max_digits": "max_digits",
        "max_length": "max_length",
        "min_length": "min_length",
        "multiple_of": "multiple_of",
        "pattern": "pattern",
    }
    # Currently unmapped attributes
    other_attributes = {
        "fail_fast": "fail_fast",
        "field_title_generator": "field_title_generator",
        "repr": "repr",
        "union_mode": "union_mode",
        "validate_default": "validate_default",
        "validation_alias": "validation_alias",
        "strict": "strict",
    }
    # Add field_info attributes
    field_info = from_model.model_fields[field_name]
    field_kwargs = {
        y: getattr(field_info, x)
        for x, y in field_info_attributes.items()
        if getattr(field_info, x) is not None
    }
    # Special case: always add default
    field_kwargs["default"] = field_info.default
    # Add field_info.metadata attributes
    for metadata in field_info.metadata:
        for x, y in metadata_attributes.items():
            if hasattr(metadata, x):
                field_kwargs[y] = getattr(metadata, x)
    # Override any attributes provided in kwargs
    field_kwargs.update(kwargs)
    # Create field
    return Field(**field_kwargs)


def set_env_variables(
    app_type: AppType | str,
    dev_idp_config: DevIdpConfig | str,
    dev_repository_config: DevRepositoryConfig | str,
    extra_settings_files: list[Path] | None = None,
    general_cfg_path: Path | None = None,
    cfg_path: Path | None = None,
) -> None:
    # Parse input
    if isinstance(app_type, str):
        if app_type.upper() in AppType.__members__:
            app_type_enum = AppType[app_type.upper()]
            app_type_str = app_type_enum.value.lower()
        else:
            app_type_str = app_type.lower()
            app_type_enum = None
    else:
        app_type_str = app_type.value.lower()
        app_type_enum = app_type
    if isinstance(dev_idp_config, str):
        dev_idp_config_enum = DevIdpConfig[dev_idp_config.upper()]
    else:
        dev_idp_config_enum = dev_idp_config
    if isinstance(dev_repository_config, str):
        dev_repository_config_enum = DevRepositoryConfig[dev_repository_config.upper()]
    else:
        dev_repository_config_enum = dev_repository_config
    # Special case: set environment variables for all apps
    if app_type_enum == AppType.ALL:
        for app2 in AppTypeSet.ALL.value:
            set_env_variables(app2, dev_idp_config_enum, dev_repository_config_enum)
        return
    elif app_type_enum == AppType.CASEDB:
        set_env_variables(
            AppType.SEQDB, dev_idp_config_enum, dev_repository_config_enum
        )
    # Initialise some
    if general_cfg_path is None:
        general_cfg_path = Path.cwd() / "config"
    if cfg_path is None:
        cfg_path = Path.cwd() / "gen_epix" / app_type_str / "config"
    envvar_prefix = app_type_str.upper() + "_"
    settings_files: list[Path] = []
    # General settings
    settings_files.append(cfg_path / "settings.toml")
    # Service secrets
    settings_files.append(cfg_path / ".example.secrets.service.toml")
    # Identity provider settings
    if dev_idp_config_enum == DevIdpConfig.IDPS:
        settings_files.append(general_cfg_path / "identity_providers.toml")
    elif dev_idp_config_enum == DevIdpConfig.MOCK:
        settings_files.append(general_cfg_path / "mock_identity_provider.toml")
    elif dev_idp_config_enum == DevIdpConfig.NONE:
        settings_files.append(general_cfg_path / "no_identity_providers.toml")
    else:
        raise ValueError(f"Unknown dev_idp_config: {dev_idp_config_enum}")
    # Repository settings
    if dev_repository_config_enum in DevRepositoryConfigSet.DICT.value:
        settings_files.append(cfg_path / "settings.repository.dict.toml")
    elif dev_repository_config_enum in DevRepositoryConfigSet.SA.value:
        settings_files.append(cfg_path / "settings.repository.sa.toml")
    else:
        raise ValueError(f"Unknown dev_repository_config: {dev_repository_config_enum}")
    # Repository secrets
    if dev_repository_config_enum == DevRepositoryConfig.DICT_DEMO:
        settings_files.append(cfg_path / ".example.secrets.repository.dict.demo.toml")
    elif dev_repository_config_enum == DevRepositoryConfig.DICT_EMPTY:
        settings_files.append(cfg_path / ".example.secrets.repository.dict.empty.toml")
    elif dev_repository_config_enum == DevRepositoryConfig.SA_SQLITE_DEMO:
        settings_files.append(
            cfg_path / ".example.secrets.repository.sa_sqlite.demo.toml"
        )
    elif dev_repository_config_enum == DevRepositoryConfig.SA_SQLITE_EMPTY:
        settings_files.append(
            cfg_path / ".example.secrets.repository.sa_sqlite.empty.toml"
        )
    elif dev_repository_config_enum == DevRepositoryConfig.SA_SQL:
        settings_files.append(cfg_path / ".example.secrets.repository.sa_sql.toml")
    else:
        raise ValueError(f"Unknown dev_repository_config: {dev_repository_config_enum}")
    # Add any extra settings files at the end
    if extra_settings_files:
        settings_files.extend(extra_settings_files)
    # Set environment variables
    os.environ[envvar_prefix + "SETTINGS_FILES"] = ",".join(
        [str(x.resolve()) for x in settings_files]
    )
    os.environ[envvar_prefix + "LOG_CONFIG_FILE"] = str(
        (cfg_path / "logging.yaml").resolve()
    )


def create_demo_data_from_repository(
    user_id: str,
    entities: list,
    dict_repository: DictRepository,
    sa_repository: SARepository,
    module_root: str,
) -> None:
    model = importlib.import_module(f"{module_root}.domain.model")
    # Delete all first in reverse order
    for entity in entities[::-1]:
        model_class = entity.model_class
        with sa_repository.uow() as sa_uow:
            sa_repository.crud(
                sa_uow,
                user_id,
                model_class,
                None,
                None,
                CrudOperation.DELETE_ALL,
            )
    for entity in entities:
        model_class = entity.model_class
        with (
            dict_repository.uow() as dict_uow,
            sa_repository.uow() as sa_uow,
        ):
            objs: list[model.Model] = dict_repository.crud(  # type: ignore[assignment]
                dict_uow,
                user_id,
                model_class,
                None,
                None,
                CrudOperation.READ_ALL,
                return_copy=False,
            )
            sa_repository.crud(
                sa_uow,
                user_id,
                model_class,
                objs,
                None,
                CrudOperation.CREATE_SOME,
            )


def load_demo_data(
    app_type: AppType,
    module_root: str,
    connect_timeout: float = 1,
    verbose: bool = True,
) -> None:
    domain: Domain = importlib.import_module(f"{module_root}.domain").DOMAIN
    # Import the sa_model module to register the models
    importlib.import_module(f"{module_root}.repositories.sa_model")
    # Get classes and config for the app type

    enum = importlib.import_module(f"{module_root}.domain.enum")

    set_env_variables(app_type, DevIdpConfig.MOCK, DevRepositoryConfig.DICT_DEMO)
    dict_app_cfg = AppCfg(
        app_type.value, enum.ServiceType, enum.RepositoryType, log_setup=False
    )
    set_env_variables(app_type, DevIdpConfig.MOCK, DevRepositoryConfig.SA_SQLITE_DEMO)
    sa_sqlite_app_cfg = AppCfg(
        app_type.value, enum.ServiceType, enum.RepositoryType, log_setup=False
    )
    set_env_variables(app_type, DevIdpConfig.MOCK, DevRepositoryConfig.SA_SQL)
    sa_sql_app_cfg = AppCfg(
        app_type.value, enum.ServiceType, enum.RepositoryType, log_setup=False
    )
    user_id = dict_app_cfg.cfg["service"]["auth"]["props"]["root"]["user"]["id"]
    for service_type in enum.ServiceType:
        # # TODO: TEMPORARY for debugging, remove later
        # if service_type.value != "CASE":
        #     continue
        dict_repository_cfg = dict_app_cfg.cfg["repository"].get(service_type.value)
        if not dict_repository_cfg:
            continue
        entities = domain.get_dag_sorted_entities(
            service_type=service_type, persistable=True
        )
        # Create dict repository, which is assumed to always be available
        dict_repository_class: type[DictRepository] = dict_repository_cfg["class"]
        demo_dict_file = Path(dict_repository_cfg["props"]["file"]).resolve()
        empty_dict_file = Path(
            str(demo_dict_file).replace(".full.", ".empty.")
        ).resolve()
        zip_file: str = str(demo_dict_file).replace(".pkl.gz", ".zip")
        start_time = datetime.datetime.now()
        dict_repository: DictRepository = (
            dict_repository_class.create_repository(  # type:ignore[assignment]
                entities=entities, file=zip_file
            )
        )
        end_time = datetime.datetime.now()
        if verbose:
            print(
                f"App {app_type.value}, service {service_type.value}: demo data parsed in {end_time - start_time}s"
            )
        # Write empty and demo dict repository to file
        start_time = datetime.datetime.now()
        with gzip.open(empty_dict_file, "wb") as handle:
            pickle.dump({x: {} for x in dict_repository.db}, handle)
        with gzip.open(demo_dict_file, "wb") as handle:
            pickle.dump(dict_repository.db, handle)
        end_time = datetime.datetime.now()
        if verbose:
            print(
                f"App {app_type.value}, service {service_type.value}: dict repository written to file in {end_time - start_time}s"
            )
        # Create empty and demo SA_SQLITE repositories
        sa_sqlite_repository_cfg = sa_sqlite_app_cfg.cfg["repository"][
            service_type.value
        ]
        sa_repository_class: type[SARepository] = sa_sqlite_repository_cfg["class"]
        demo_sa_sqlite_file = Path(sa_sqlite_repository_cfg["props"]["file"]).resolve()
        empty_sa_sqlite_file = Path(
            str(demo_sa_sqlite_file).replace(".full", ".empty")
        ).resolve()
        start_time = datetime.datetime.now()
        # Empty repository
        sa_repository_class.create_repository(
            entities=entities,
            file=empty_sa_sqlite_file,
            name=service_type.value,
            recreate_sqlite_file=True,
        )
        # Full repository
        sa_sqlite_repository: SARepository = (
            sa_repository_class.create_repository(  # type:ignore[assignment]
                entities=entities,
                file=demo_sa_sqlite_file,
                name=service_type.value,
                recreate_sqlite_file=True,
            )
        )
        create_demo_data_from_repository(
            user_id, entities, dict_repository, sa_sqlite_repository, module_root
        )
        end_time = datetime.datetime.now()
        if verbose:
            print(
                f"App {app_type.value}, service {service_type.value}: sa_sqlite repository written to file in {end_time - start_time}s"
            )
        # Create empty SA_SQL repository or loaded with demo data
        sa_sql_repository_cfg = sa_sql_app_cfg.cfg["repository"][service_type.value]
        connection_string = sa_sql_repository_cfg["props"]["connection_string"]
        if "mssql" in connection_string:
            connect_args = {
                "timeout": connect_timeout,
                "login_timeout": connect_timeout,
            }
        elif "pyodcb" in connection_string:
            connect_args = {
                "connect_timeout": connect_timeout,
                "timeout": connect_timeout,
            }
        else:
            connect_args = {}
        # Skip load if no connection can be made
        if exception := sa_repository_class.test_connection(
            connection_string, **connect_args
        ):
            if verbose:
                print(
                    # f"App {app_type.value}, service {service_type.value}: sa_sql connection failed: {exception}"
                    f"App {app_type.value}, service {service_type.value}: sa_sql connection failed"
                )
            continue
        start_time = datetime.datetime.now()
        sa_sql_repository: SARepository = (
            sa_repository_class.create_repository(  # type:ignore[assignment]
                entities=entities,
                connection_string=connection_string,
                name=service_type.value,
            )
        )
        create_demo_data_from_repository(
            user_id, entities, dict_repository, sa_sql_repository, module_root
        )
        end_time = datetime.datetime.now()
        if verbose:
            print(
                f"App {app_type.value}, service {service_type.value}: sa_sql repository loaded in {end_time - start_time}s"
            )


def get_app_cfgs(
    app_type: AppType,
    service_type_enum: type[Enum],
    repository_type_enum: type[Enum],
    test_type: Enum | str,
    dev_idp_config: DevIdpConfig = DevIdpConfig.NONE,
    general_cfg_path: Path | None = None,
    cfg_path: Path | None = None,
    extra_settings_files: (
        list[Path | str] | Path | str | None
    ) = "./test/test_client/settings.toml",
    seqdb_app_cfgs: dict[str, AppCfg] | None = None,
    log_setup: bool = False,
    log_level: str | int = logging.ERROR,
) -> dict[str, AppCfg]:
    """
    Create all casedb and seqdb app cfgs with a name for the given test type and
    dev repository config so that they can be reused in tests
    """
    if isinstance(test_type, Enum):
        test_type = test_type.value
    if extra_settings_files:
        if not isinstance(extra_settings_files, list):
            extra_settings_files = [extra_settings_files]
        for i, file in enumerate(extra_settings_files):
            if not isinstance(file, (str, Path)):
                raise ValueError("extra_settings_files must be a list of str or Path")
            if isinstance(file, str):
                file = Path(file)
            if not file.is_file():
                raise ValueError(
                    f"extra_settings_file {file} does not exist or is not a file"
                )
            extra_settings_files[i] = file.resolve()
    app_cfgs: dict[str, AppCfg] = {}
    for dev_repository_config in DevRepositoryConfig:
        name = f"{test_type}__{dev_repository_config.value}"
        set_env_variables(
            app_type,
            dev_idp_config,
            dev_repository_config,
            general_cfg_path=general_cfg_path,
            cfg_path=cfg_path,
            extra_settings_files=extra_settings_files,
        )
        app_cfgs[name] = AppCfg(
            app_type,
            service_type_enum,
            repository_type_enum,
            name=name,
            log_setup=log_setup,
        )
        # Set log level
        app_cfgs[name].set_log_level(log_level)
        # Add seqdb app_cfg to casedb app_cfg for seqdb service local app so that when the latter is instantiated, it can directly use this app_cfg without risk of having seqdb env variables being altered in the meantime
        if app_type == AppType.CASEDB and seqdb_app_cfgs is not None:
            app_cfgs[name].cfg["service"]["seqdb"]["props"]["seqdb_local_app"][
                "app_cfg"
            ] = seqdb_app_cfgs[name]
    return app_cfgs



from collections.abc import Iterable
from typing import Any


class DomainException(Exception):
    def __init__(self, message: str | None):
        self.message = message


class DataException(DomainException):
    def __init__(self, message: str | None, ids: Iterable | None = None):
        super().__init__(message)
        self.ids = ids


class InvalidArgumentsError(DataException):
    def __init__(self, message: str, ids: Iterable | None = None):
        super().__init__(message, ids=ids)


class IdsError(DataException):
    def __init__(self, message: str, ids: Iterable | None = None):
        super().__init__(message, ids=ids)


class InvalidIdsError(IdsError):
    def __init__(self, message: str, ids: Iterable | None = None):
        super().__init__(message, ids=ids)


class DuplicateIdsError(IdsError):
    def __init__(self, message: str, ids: Iterable | None = None):
        super().__init__(message, ids=ids)


class InvalidModelIdsError(IdsError):
    def __init__(self, message: str, ids: Iterable | None = None):
        super().__init__(message, ids=ids)


class AlreadyExistingIdsError(IdsError):
    def __init__(self, message: str, ids: Iterable | None = None):
        super().__init__(message, ids=ids)


class InvalidLinkIdsError(IdsError):
    def __init__(self, message: str, ids: Iterable | None = None):
        super().__init__(message, ids=ids)


class LinkConstraintViolationError(IdsError):
    def __init__(
        self,
        message: str,
        ids: Iterable | None = None,
        linked_ids: Iterable | None = None,
    ):
        super().__init__(message, ids=ids)
        self.linked_ids = linked_ids


class UniqueConstraintViolationError(DataException):
    def __init__(
        self,
        message: str,
        ids: Iterable | None = None,
        duplicate_key_ids: Iterable | None = None,
    ):
        super().__init__(message, ids=ids)
        self.duplicate_key_ids = duplicate_key_ids


class NotNullConstraintViolationError(DataException):
    def __init__(
        self,
        message: str,
        ids: Iterable | None = None,
        column_names: Iterable[str] | None = None,
    ):
        super().__init__(message, ids=ids)
        self.column_names = column_names


class NoResultsError(DataException):
    def __init__(self, message: str | None = None):
        # Message is optional
        super().__init__(message)


class ServiceException(DomainException):
    def __init__(
        self, message: str | None = None, http_props: dict[str, Any] | None = None
    ):
        # Message is optional
        super().__init__(message)
        if http_props is None:
            http_props = {}
        self._init_http_props(http_props, 500)

    def get_http_status_code(self) -> int:
        return int(self.http_props["status_code"])

    def get_http_other_props(self) -> dict[str, Any]:
        return {x: y for x, y in self.http_props.items() if x not in {"status_code"}}

    def _init_message(self, message: str | None, default_message: str) -> None:
        super().__init__(message=default_message if not message else message)

    def _init_http_props(
        self, http_props: dict[str, Any], http_status_code: int
    ) -> None:
        self.http_props = {**http_props}
        self.http_props["status_code"] = self.http_props.get(
            "status_code", http_status_code
        )


class InitializationServiceError(ServiceException):
    pass


class RepositoryInitializationServiceError(InitializationServiceError):
    pass


class RepositoryServiceError(ServiceException):
    pass


class AuthException(ServiceException):
    pass


class CredentialsAuthError(AuthException):
    def __init__(
        self, message: str | None = None, http_props: dict[str, Any] | None = None
    ):
        super().__init__(message, http_props)
        if http_props is None:
            http_props = {}
        self._init_message(message, "Could not validate credentials")
        self._init_http_props(http_props, 401)


class UnauthorizedAuthError(AuthException):
    def __init__(
        self, message: str | None = None, http_props: dict[str, Any] | None = None
    ):
        super().__init__(message, http_props)
        if http_props is None:
            http_props = {}
        self._init_message(message, "Unauthorized: Invalid credentials")
        self._init_http_props(http_props, 403)


class UserNotFoundAuthError(AuthException):
    def __init__(
        self, message: str | None = None, http_props: dict[str, Any] | None = None
    ):
        super().__init__(message, http_props)
        if http_props is None:
            http_props = {}
        self._init_message(message, "User not found")
        self._init_http_props(http_props, 404)


class UserAlreadyExistsAuthError(AuthException):
    def __init__(
        self, message: str | None = None, http_props: dict[str, Any] | None = None
    ):
        super().__init__(message, http_props)
        if http_props is None:
            http_props = {}
        self._init_message(message, "User already exists")
        self._init_http_props(http_props, 409)


class ServiceUnavailableError(ServiceException):
    def __init__(
        self, message: str | None = None, http_props: dict[str, Any] | None = None
    ):
        super().__init__(message, http_props)
        if http_props is None:
            http_props = {}
        self._init_message(message, "Service unavailable")
        self._init_http_props(http_props, 503)



from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.service import BaseGeoService


class GeoService(BaseGeoService):
    def retrieve_containing_region(
        self, _cmd: command.RetrieveContainingRegionCommand
    ) -> list[model.Region | None]:
        raise NotImplementedError()
